
static char rcsid[] =
	"$Id: message_passing.c,v 1.22 1997/07/22 19:24:59 pvmsrc Exp $";

/***********************/
/*  message_passing.c  */
/*  tests for:         */
/*     head-head       */
/*     triangle        */
/*     funneling       */
/***********************/

#include "pvm_test.h"
#include "globals.h"

/******************/
/*  HEAD TO HEAD  */
/******************/

/*  MASTER  */

int
head_head( test_num, host, is_empty, send_mode, recv_mode,
		route, data_format )
int		test_num;
int		host;
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	struct timeval t1, t2;

	double	Total_time = 0;
	double	Average_time;

	int		*outbuf;
	int		*inbuf;

	int		result;
	int		count;
	int		info;
	int		mlen;
	int		ack;
	int		tid;
	int		i;

	inbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return( FAILED );

	outbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return( FAILED );

	/** Check if options possible **/

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	if ( vflag )
	{
		errprintf( "%s (%s) < --- > %s (%s)\n", 
				hostpool[0], hostarch[0],
				hostpool[ host ], hostarch[ host ] );
	}

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS, 
			PvmTaskHost, hostpool[ host ], 1, &tid ) < 0 )
	{
		return( FAILED );
	}

	result = FAILED;

	if ( master_send( &tid, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) &messages_length, 1, PVM_INT,
			hostpool[ host ] ) < 0 )
		goto fail;

	ack = timeout_wait_message( tid, ACKNOWLEDGE, host, VERBOSE );

	if ( ack < 0 )
		goto fail;

	if ( master_recv( tid, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL, hostpool[ host ] ) < 0 )
		goto fail;

	info = pvm_setopt( PvmRoute, route );
	CKERR( info, "pvm_setopt", NOHOST, goto fail );

	mlen = ( is_empty == EMPTY ) ? 0 : messages_length;

	if ( timing_total || timing_average )
		gettimeofday( &t1, NULL );

	for( i=0 ; i < num_messages ; i++ )
	{
		if ( master_send( &tid, 1, HEAD_HEAD, send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT,
				hostpool[ host ] ) < 0 )
			goto fail;

		for ( count=0 ; count < BUSY_LOOP_PERF ; count++ )
			if ( pvm_probe( tid, HEAD_HEAD ) > 0 )
				break;

		if ( count == BUSY_LOOP_PERF )
		{
			errprintf( "\tTimeout expired from host %s\n", 
					hostpool[host] );
			goto fail;
		}

		/* ack = timeout_wait_message( tid, HEAD_HEAD, host, */
				/* VERBOSE ); */

		/* if ( ack < 0 ) */
			/* goto fail; */

		if ( master_recv( tid, HEAD_HEAD, recv_mode,
				(void *) inbuf, mlen, PVM_INT, (int *) NULL,
				hostpool[ host ] ) < 0 )
			goto fail;
	}

	if ( timing_total || timing_average )
		gettimeofday( &t2, NULL );

	if ( timing_total )
	{
		Total_time = TIME_ELAPSED;

		if ( vflag )
			errprintf( "\tTotal time = %f sec.\n", Total_time );
	}

	if ( timing_average )
	{
		Average_time = TIME_ELAPSED / ( (double) num_messages );

		if ( vflag )
			errprintf( "\tAverage time = %f sec.\n", Average_time );
	}

	result = PASSED;

fail:

	pvm_setopt( PvmRoute, PvmAllowDirect );

	pvm_kill( tid );

	return( result );
}


/*  SLAVE  */

void
head_head_slave( is_empty, send_mode, recv_mode, route, data_format )
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	int		*outbuf;
	int		*inbuf;

	int		length;
	int		mlen;

	slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
			(void *) &length, 1, PVM_INT );

	inbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return;

	outbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return;

	mlen = ( is_empty == EMPTY ) ? 0 : length;

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	pvm_setopt( PvmRoute, route );

	while ( 1 )
	{
		slave_recv( MYPARENT, HEAD_HEAD, recv_mode,
				(void *) inbuf, mlen, PVM_INT );

		slave_send( &MYPARENT, 1, HEAD_HEAD, send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT );
	}
}


/**************/
/*  TRIANGLE  */
/**************/

/*  MASTER  */

int
triangle( test_num, host1, host2,
		is_empty, send_mode, recv_mode, route, data_format )
int		test_num;
int		host1;
int		host2;
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	struct timeval t1, t2;

	double	Total_time = 0;
	double	Average_time;

	int		*outbuf;
	int		*inbuf;

	int		you_are_the_second = 2;
	int		you_are_the_first = 1;
	int		result;
	int		count;
	int		info;
	int		mlen;
	int		tid1;
	int		tid2;
	int		ack;
	int		i;

	inbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return( FAILED );

	outbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return( FAILED );

	/** Check if options possible **/

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	if ( vflag )
	{
		errprintf( "%s (%s) --- > %s (%s) --- > %s (%s) --- > \n", 
				hostpool[0], hostarch[0], 
				hostpool[host1], hostarch[host1], 
				hostpool[host2], hostarch[host2] );
	}

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS, 
			PvmTaskHost, hostpool[ host1 ], 1, &tid1 ) < 0 )
	{
		return( FAILED );
	}

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS, 
			PvmTaskHost, hostpool[ host2 ], 1, &tid2 ) < 0 )
	{
		pvm_kill( tid1 );
		return( FAILED );
	}

	result = FAILED;

	if ( triangle_init_slave( tid1, tid2, send_mode, data_format,
			1, hostpool[ host1 ] ) < 0 )
		goto fail;

	if ( triangle_init_slave( tid2, tid1, send_mode, data_format,
			2, hostpool[ host2 ] ) < 0 )
		goto fail;

	/* Acknoledgements */

	ack = timeout_wait_message( tid1, ACKNOWLEDGE, host1, VERBOSE );

	if ( ack < 0 )
		goto fail;

	if ( master_recv( tid1, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL, hostpool[ host1 ] ) < 0 )
		goto fail;

	ack = timeout_wait_message( tid2, ACKNOWLEDGE, host2, VERBOSE );

	if ( ack < 0 )
		goto fail;

	if ( master_recv( tid2, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL, hostpool[ host2 ] ) < 0 )
		goto fail;

	/*	Test Body */

	info = pvm_setopt( PvmRoute, route );
	CKERR( info, "pvm_setopt", NOHOST, goto fail );

	mlen = ( is_empty == EMPTY ) ? 0 : messages_length;

	if ( timing_total || timing_average )
		gettimeofday( &t1, NULL );

	for( i=0 ; i < num_messages ; i++ )
	{
		if ( master_send( &tid1, 1, TRIANGLE, send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT,
				hostpool[ host1 ] ) < 0 )
			goto fail;

		for ( count=0 ; count < BUSY_LOOP_PERF ; count++ )
			if ( pvm_probe( tid2, TRIANGLE ) > 0 )
				break;

		if ( count == BUSY_LOOP_PERF )
		{
			errprintf( "\tTimeout expired from host %s\n", 
					hostpool[host2] );
			goto fail;
		}

		/* ack = timeout_wait_message( tid2, TRIANGLE, host2, */
				/* VERBOSE ); */

		/* if ( ack < 0 ) */
			/* goto fail; */

		if ( master_recv( tid2, TRIANGLE, recv_mode,
				(void *) inbuf, mlen, PVM_INT, (int *) NULL,
				hostpool[ host2 ] ) < 0 )
			goto fail;
	}

	if ( timing_total || timing_average )
		gettimeofday( &t2, NULL );

	if ( timing_total )
	{
		Total_time = TIME_ELAPSED;

		if ( vflag )
			errprintf( "\tTotal time = %f sec.\n", Total_time );
	}

	if ( timing_average )
	{
		Average_time = TIME_ELAPSED / ( (double) num_messages );

		if ( vflag )
			errprintf( "\tAverage time = %f sec.\n", Average_time );
	}

	result = PASSED;

fail:

	pvm_kill( tid1 );
	pvm_kill( tid2 );

	pvm_setopt( PvmRoute, PvmAllowDirect );

	return( result );
}


int
triangle_init_slave( tid, other_tid, send_mode, data_format,
		index, host )
int		tid;
int		other_tid;
int		data_format;
int		index;
char	*host;
{
	int		init[3];

	int		info;

	init[0] = other_tid;
	init[1] = index;
	init[2] = messages_length;

	if ( (info = master_send( &tid, 1, ACKNOWLEDGE,
			send_mode, data_format,
			(void *) init, 3, PVM_INT, host )) < 0 )
		return( info );

	return( PASSED );
}


/*  SLAVE  */

void
triangle_slave( is_empty, send_mode, recv_mode, route, data_format )
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	int		init[3];

	int		*outbuf;
	int		*inbuf;

	int		tidother;
	int		length;
	int		mlen;
	int		rank;
	int		from;
	int		to;

	slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
			(void *) init, 3, PVM_INT );

	tidother	=	init[0];
	rank		=	init[1];
	length		=	init[2];

	inbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return;

	outbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return;

	mlen = ( is_empty == EMPTY ) ? 0 : length;

	if ( rank == 1 )
	{
		to = tidother;
		from = pvm_parent();
	}

	else
	{
		to = pvm_parent();
		from = tidother;
	}

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	pvm_setopt( PvmRoute, route );

	while ( 1 )
	{
		slave_recv( from, TRIANGLE, recv_mode,
				(void *) inbuf, mlen, PVM_INT );

		slave_send( &to, 1, TRIANGLE, send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT );
	}
}


/***************/
/*  FUNNELING  */
/***************/

/*  MASTER  */

int
funneling( test_num, array_of_index, length,
		is_empty, send_mode, recv_mode, route, data_format )
int		test_num;
int		*array_of_index;
int		length;
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	struct timeval t1, t2;

	double	Total_time = 0;
	double	Average_time;

	int		*effective_array_of_index;
	int		*outbuf;
	int		*inbuf;
	int		*tid;

	int		number_of_process;
	int		result;
	int		count;
	int		host;
	int		info;
	int		mlen;
	int		i, j;
	int		ack;

	inbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return( FAILED );

	outbuf = (int *) malloc( sizeof( int ) * messages_length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return( FAILED );

	/** Check if options possible **/

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	number_of_process = module_list[ test_num ].flag.num_req_hosts;

	effective_array_of_index = (int *) malloc( number_of_process
			* sizeof( int ) );
	if ( !memcheck( effective_array_of_index, "Host Index Array" ) )
		return( FAILED );

	tid = (int *) malloc( number_of_process * sizeof( int ) );
	if ( !memcheck( tid, "TID Array" ) )
		return( FAILED );

	for ( i=0 ; i < length ; i++ )
		effective_array_of_index[i] = array_of_index[i];

	for ( i=length ; i < number_of_process ; i++ )
		effective_array_of_index[i] = array_of_index[ i % length ];

	if ( vflag )
	{
		errprintf( "%s (%s)\n", hostpool[0], hostarch[0] );

		for ( i=1 ; i < number_of_process ; i++ )
		{
			errprintf( "\t\t --- > %s (%s)\n", 
					hostpool[ effective_array_of_index[i] ], 
					hostarch[ effective_array_of_index[i] ] );
		}
	}

	result = FAILED;

	for ( i=1 ; i < number_of_process ; i++ )
	{
		host = effective_array_of_index[i];

		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS, 
				PvmTaskHost, hostpool[ host ], 1, &( tid[i] ) ) < 0 )
		{
			goto fail;
		}

		if ( master_send( &(tid[i]), 1, ACKNOWLEDGE,
				send_mode, data_format,
				(void *) &messages_length, 1, PVM_INT,
				hostpool[ host ] ) < 0 )
			goto fail;

		ack = timeout_wait_message( tid[i], ACKNOWLEDGE,
				host, VERBOSE );

		if ( ack < 0 )
			goto fail;

		if ( master_recv( tid[i], ACKNOWLEDGE, recv_mode,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ host ] ) < 0 )
			goto fail;
	}

	info = pvm_setopt( PvmRoute, route );
	CKERR( info, "pvm_setopt", NOHOST, goto fail );

	mlen = ( is_empty == EMPTY ) ? 0 : messages_length;

	if ( timing_total || timing_average )
		gettimeofday( &t1, NULL );

	for ( j=0 ; j < num_messages ; j++ )
	{
		if ( master_send( tid + 1, number_of_process - 1, FUNNELING,
				send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT,
				NOHOST ) < 0 )
			goto fail;

		for ( i=1 ; i < number_of_process ; i++ )
		{
			for ( count=0 ; count < BUSY_LOOP_PERF ; count++ )
				if ( pvm_probe( -1, FUNNELING ) > 0 )
					break;

			if ( count == BUSY_LOOP_PERF )
			{
				errprintf( "\tTimeout expired\n" );
				goto fail;
			}

			/* ack = timeout_wait_message( -1, FUNNELING, -1, */
					/* VERBOSE ); */

			/* if ( ack < 0 ) */
				/* goto fail; */

			if ( master_recv( -1, FUNNELING, recv_mode,
					(void *) inbuf, mlen, PVM_INT, (int *) NULL,
					NOHOST ) < 0 )
				goto fail;
		}
	}

	if ( timing_total || timing_average )
		gettimeofday( &t2, NULL );

	if ( timing_total )
	{
		Total_time = TIME_ELAPSED;

		if ( vflag )
			errprintf( "\tTotal time = %f sec.\n", Total_time );
	}

	if ( timing_average )
	{
		Average_time = TIME_ELAPSED / ( (double) num_messages );

		if ( vflag )
			errprintf( "\tAverage time = %f sec.\n", Average_time );
	}

	result = PASSED;

fail:

	for ( i=1 ; i < number_of_process ; i++ )
		pvm_kill( tid[i] );

	pvm_setopt( PvmRoute, PvmAllowDirect );

	free( effective_array_of_index );
	free( tid );

	return( result );
}


/*  SLAVE  */

void
funneling_slave( is_empty, send_mode, recv_mode, route, data_format )
int		is_empty;
int		send_mode;
int		recv_mode;
int		route;
int		data_format;
{
	int		*outbuf;
	int		*inbuf;

	int		length;
	int		mlen;

	slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
			(void *) &length, 1, PVM_INT );

	inbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( inbuf, "Messaging Input Buffer" ) )
		return;

	outbuf = (int *) malloc( sizeof( int ) * length );
	if ( !memcheck( outbuf, "Messaging Output Buffer" ) )
		return;

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	pvm_setopt( PvmRoute, route );

	mlen = ( is_empty == EMPTY ) ? 0 : length;

	while ( 1 )
	{
		slave_recv( MYPARENT, FUNNELING, recv_mode,
				(void *) inbuf, mlen, PVM_INT );

		slave_send( &MYPARENT, 1, FUNNELING, send_mode, data_format,
				(void *) outbuf, mlen, PVM_INT );
	}
}

