
static char rcsid[] =
	"$Id: test_list.c,v 1.44 2001/02/05 22:58:22 pvmsrc Exp $";

/*******************/
/*  test_list.c    */
/*  tests masters  */
/*******************/

#include "pvm_test.h"
#include "globals.h"

/***********************/
/*  pvm_mstat correct  */
/***********************/

int
test_mstat_correct( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i=0;
	int		mstat;
	int		result = PASSED;

	for ( i=1 ; i < length ; i++ )
	{
		mstat = pvm_mstat( hostpool[ array_of_index[i] ] );

		CKERR( mstat, "pvm_mstat", hostpool[ array_of_index[i] ],
				result = FAILED );
	}

	return( result );
}


/*************************/
/*  pvm_mstat incorrect  */
/*************************/

int
test_mstat_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		result = PASSED;
	int		mstat;

	for ( i=1 ; i < length ; i++ )
	{
		mstat = pvm_mstat( hostpool[ array_of_index[i] ] );

		if ( mstat == PvmOk )
		{
			errprintf( "\tpvm_stat(): Detect a not added host (%s)\n",
					hostpool[ array_of_index[i] ] );
			result = FAILED;
			continue;
		}

		CKERR( mstat, "pvm_mstat", hostpool[ array_of_index[i] ],
				result = FAILED );
	}

	return( result );
}


/**************************/
/*  pvm_addhosts correct  */
/**************************/

int
test_addhosts_correct( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		stat, mstat;
	int		info;
	int		result = PASSED;

	i = 1;

	while( hostpool[i] != NULL )
	{
		stat = pvm_addhosts( hostpool + i, 1, &info );

		CKERR( stat, "pvm_addhosts", hostpool[i],
				{ result = FAILED; continue; } );

		info = wait_number_hosts( i + 1 );

		if ( info < 0 )
			return( FAILED );

		mstat = pvm_mstat( hostpool[i] );

		CKERR( mstat, "pvm_mstat", hostpool[i], result = FAILED );

		i++;
	}

	return( result );
}


/****************************/
/*  pvm_addhosts incorrect  */
/****************************/

int
test_addhosts_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		stat, mstat;
	int		info;
	int		result = PASSED;

	i = 1;

	while( hostpool[i] != NULL )
	{
		stat = pvm_addhosts( hostpool + i, 1, &info );

		CKERR( stat, "pvm_addhosts", hostpool[i],
				{ result = FAILED; continue; } );

		info = wait_number_hosts( i + 1 );

		if ( info < 0 )
			return( FAILED );

		mstat = pvm_mstat( hostpool[i] );

		CKERR( mstat, "pvm_mstat", hostpool[i], result = FAILED );

		i++;
	}

	if ( result == FAILED )
	{
		errprintf( "\tpvm_addhosts() should not have failed yet\n" );
		return( FAILED );
	}

	i=0;

	while( hostpool[i] != 0 )
	{
		stat = pvm_addhosts( hostpool + i, 1, &info );

		EXPECTED( "pvm_addhosts", PvmDupHost, info, hostpool[i],
				result = FAILED );

		i++;
	}

	return( result );
}


/**************************/
/*  pvm_delhosts correct  */
/**************************/

int
test_delhosts_correct( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i=0;
	int		stat, mstat;
	int		info;
	int		result = PASSED;

	for ( i=1 ; i < length ; i++ )
	{
		stat = pvm_delhosts( &hostpool[ array_of_index[i] ], 1, &info );

		CKERR( stat, "pvm_delhosts", hostpool[ array_of_index[i] ],
				{ result = FAILED; continue; } );

		info = wait_number_hosts( length - i );

		if ( info < 0 )
			return( FAILED );

		mstat = pvm_mstat( hostpool[ array_of_index[i] ] );

		if ( mstat != PvmNoHost && mstat != PvmHostFail )
		{
			CKERR( mstat, "pvm_mstat", hostpool[ array_of_index[i] ],
					{ result = FAILED; continue; } );
		}
	}

	return( result );
}


/****************************/
/*  pvm_delhosts incorrect  */
/****************************/

int
test_delhosts_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		stat, mstat;
	int		info;
	int		result = PASSED;

	for ( i=1 ; i < length ; i++ )
	{
		stat = pvm_delhosts( &hostpool[i], 1, &info );

		CKERR( stat, "pvm_delhosts", hostpool[i],
				{ result = FAILED; continue; } );

		info = wait_number_hosts( length - i );

		if ( info < 0 )
			return( FAILED );

		mstat = pvm_mstat( hostpool[i] );

		if ( mstat != PvmNoHost && mstat != PvmHostFail )
		{
			CKERR( mstat, "pvm_mstat", hostpool[i],
				{ result = FAILED; continue; } );
		}
	}

	if ( result == FAILED )
	{
		errprintf( "\tpvm_delhosts(): should not have failed yet\n" );
		return( FAILED );
	}

	for ( i=0 ; i < length ; i++ )
	{
		stat = pvm_delhosts( &hostpool[i], 1, &info );

		if ( stat >= 1 )
		{
			errprintf(
			"\tpvm_delhosts():  expected negative value for host %s\n",
				hostpool[i] );
			result = FAILED;
		}
	}

	return( result );
}


/**********************************/
/*  Race pvm_config pvm_addhosts  */
/**********************************/

int
test_race_addhosts_config( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct pvmhostinfo *hostp;

	int		i;
	int		result = PASSED;
	int		stat;
	int		*infos;
	int		num_found;
	int		narch;

	i = 1;

	while( hostpool[i++] != NULL );

	length = i - 2;

	if ( length == 0 )
		return( PASSED );

	infos = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( infos, "Info Array" ) )
		return( FAILED );

	stat = pvm_addhosts( hostpool + 1, length, infos );

	if ( stat < length )
	{
		for ( i=0 ; i < length ; i++ )
			CKERR( infos[i], "pvm_addhosts", hostpool[i], NOHANDLER );

		result = FAILED;

		goto fail;
	}

	sleep( SNOOZ );

	stat = pvm_config( &num_found, &narch, &hostp );

	CKERR( stat, "pvm_config", NOHOST,
			{ result = FAILED; goto fail; } );

	if ( num_found != length + 1 )
	{
		errprintf(
			"\tpvm_config(): Didn't get up-to-date info in time\n" );
		result = FAILED;
	}

fail:

	free( infos );

	return( result );
}


/**********************************/
/*  Race pvm_config pvm_delhosts  */
/**********************************/

int
test_race_delhosts_config( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct pvmhostinfo *hostp;

	int		i;
	int		stat;
	int		*infos;
	int		result = PASSED;
	int		num_found;
	int		narch;

	if ( length == 1 )
		return( PASSED );

	infos = (int *) malloc( ( length - 1 ) * sizeof( int ) );
	if ( !memcheck( infos, "Info Array" ) )
		return( FAILED );

	stat = pvm_delhosts( hostpool + 1, length - 1, infos );

	if ( stat < length - 1 )
	{
		for ( i=0 ; i < length - 1 ; i++ )
		{
			CKERR( infos[i], "pvm_delhosts", hostpool[ i + 1 ],
					NOHANDLER );
		}

		result = FAILED;

		goto fail;
	}

	sleep( SNOOZ );

	stat = pvm_config( &num_found, &narch, &hostp );

	CKERR( stat, "pvm_config", NOHOST,
			{ result = FAILED; goto fail; } );

	if ( num_found != 1 )
	{
		errprintf(
			"\tpvm_config(): Didn't get up-to-date info in time\n" );
		result = FAILED;
	}

fail:

	wait_number_hosts( 1 );

	free( infos );

	return( result );
}


/********************/
/*  pvm_hostsync()  */
/********************/

int
test_hostsync( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		info;
	int		result = PASSED;
	struct pvmhostinfo *hostp;
	struct timeval clk, delta;
	int		nhost;
	int		narch;

	if ( length == 1 )
		return( PASSED );

	info = pvm_config( &nhost, &narch, &hostp );

	CKERR( info, "pvm_config", NOHOST, return( FAILED ) );

	for ( i=1 ; i < length ; i++ )
	{
		info = pvm_hostsync( hostp[i].hi_tid, &clk, &delta );

		CKERR( info, "pvm_hostsync", hostpool[ array_of_index[i] ],
				result = FAILED );
	}

	return( result );
}


/***********************/
/*  pvm_pstat correct  */
/***********************/

int
test_pstat_correct( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		mytid;
	int		pstat;

	mytid = pvm_mytid();

	CKERR( mytid, "pvm_mytid", NOHOST, return( FAILED ) );

	pstat = pvm_pstat( mytid );

	CKERR( pstat, "pvm_pstat", NOHOST, return( FAILED ) );

	return( PASSED );
}


/*************************/
/*  pvm_pstat incorrect  */
/*************************/

int
test_pstat_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		mytid;
	int		pstat;

	mytid = pvm_mytid();

	CKERR( mytid, "pvm_mytid", NOHOST, return( FAILED ) );

	pstat = pvm_pstat( mytid + 3 );

	if ( pstat == PvmOk )
	{
		errprintf( "\tpvm_stat():  Recognized a non-existent task!\n" );
		return( FAILED );
	}

	else if ( pstat == PvmBadParam )
	{
		errprintf(
			"\tpvm_stat():  valid parameter was considered invalid\n" );
		return( FAILED );
	}

	else if ( pstat != PvmNoTask )
	{
		CKERR( pstat, "pvm_pstat", NOHOST, return( FAILED ) );
	}

	return( PASSED );
}


/**************************/
/*  pvm_spawn sequential  */
/**************************/

int
test_spawn( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		pstat;
	int		tid;
	int		result = PASSED;

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;

		pstat = pvm_pstat( tid );

		CKERR( pstat, "pvm_pstat", hostpool[ array_of_index[i] ],
				result = FAILED );

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/***************************/
/*  pvm_spawn simulateous  */
/***************************/

int
test_spawn_simul( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		info;
	int		ack;
	int		pstat;
	int		*tids;
	int		result = PASSED;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	if ( (info = spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskDefault, NOHOST, length, tids ))
			< 0 )
	{
		result = FAILED;
		goto fail;
	}

	for( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( -1, ACKNOWLEDGE, -1, VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( -1, ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
			result = FAILED;
	}

	for ( i=0 ; i < length ; i++ )
	{
		pstat = pvm_pstat( tids[i] );

		CKERR( pstat, "pvm_pstat", hostpool[ array_of_index[i] ],
				result = FAILED );

		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

fail:

	free( tids );

	return( result );
}


/**********************/
/*  pvm_kill correct  */
/**********************/

int
test_kill_correct( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		pstat;
	int		tid;
	int		result = PASSED;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( -1, ACKNOWLEDGE, array_of_index[i],
				VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;

			pvm_kill( tid );
		}

		else
		{
			pstat = pvm_pstat( tid );

			CKERR( pstat, "pvm_pstat", hostpool[ array_of_index[i] ],
					result = FAILED );

			if ( master_recv( -1, ACKNOWLEDGE, RECV,
					(void *) NULL, 0, 0, (int *) NULL,
					hostpool[ array_of_index[i] ] ) < 0 )
				result = FAILED;
		}

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", hostpool[ array_of_index[i] ],
				result = FAILED );
	}

	return( result );
}


/************************/
/*  pvm_kill incorrect  */
/************************/

int
test_kill_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		pstat;
	int		tid;
	int		result = PASSED;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( -1, ACKNOWLEDGE, array_of_index[i],
				VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		pstat = pvm_pstat( tid );

		CKERR( pstat, "pvm_pstat", hostpool[ array_of_index[i] ],
				result = FAILED );

		if ( master_recv( -1, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", hostpool[ array_of_index[i] ],
				result = FAILED );

		info = pvm_kill( tid );

		if ( info > 0 )
		{
			result = FAILED;

			errprintf( "\tpvm_kill():  killed a non-existent slave " );
			errprintf( "without error on host %s\n",
					hostpool[ array_of_index[i] ] );
		}
	}

	return( result );
}


/*********************/
/*  pvm_tidtohost()  */
/*********************/

int
test_tidtohost( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		tid;
	int		result = PASSED;
	struct pvmhostinfo *hostp;
	int		dtid;
	int		pstat;
	int		narch;
	int		nhost;

	info = pvm_config( &nhost, &narch, &hostp );

	CKERR( info, "pvm_config", NOHOST, return( FAILED ) );

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;

		pstat = pvm_pstat( tid );

		CKERR( pstat, "pvm_pstat", hostpool[ array_of_index[i] ],
				result = FAILED );

		dtid = pvm_tidtohost( tid );

		CKERR( dtid, "pvm_tidtohost", hostpool[ array_of_index[i] ],
				result = FAILED );

		if ( dtid >= 0 )
		{
			if ( dtid != hostp[i].hi_tid )
			{
				errprintf(
				"\tpvm_tidtohost(): incorrect host tid for host %s\n",
					hostpool[ array_of_index[i] ] );

				result = FAILED;
			}
		}

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/******************/
/*  pvm_parent()  */
/******************/

int
test_parent( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		tid;
	int		result = PASSED;
	int		bufid;
	int		response;
	int		mytid;

	mytid = pvm_mytid();

	CKERR( mytid, "pvm_mytid", NOHOST, return( FAILED ) );

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV,
				(void *) &response, 1, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		EXPECTED( "pvm_parent", mytid, response,
				hostpool[ array_of_index[i] ], result = FAILED );

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/*******************/
/*  pvm_sendsig()  */
/*******************/

int
test_sendsig( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		tid;
	int		result = PASSED;

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		ack = timeout_wait_message( tid, ACKNOWLEDGE, i, VERBOSE );

		if ( ack < 0 )
		{
	 		result = FAILED;
	 		pvm_kill( tid );
	 		continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		info = pvm_sendsig( tid, SIGINT );

		CKERR( info, "pvm_sendsig", hostpool[ array_of_index[i] ],
				{ result = FAILED; pvm_kill( tid ); continue; } );

		ack = timeout_wait_message( tid, SIGNAL, i, VERBOSE );

		if ( ack < 0 )
		{
	 		result = FAILED;
	 		pvm_kill( tid );
	 		continue;
		}

		if ( master_recv( tid, SIGNAL, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/***************/
/*  pvm_tasks  */
/***************/

int
test_tasks( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		info;
	int		ack;
	int		*tids;
	int		result = PASSED;
	struct pvmtaskinfo *taskp;
	int		ntask, orig_ntask;

	info = pvm_tasks( 0, &ntask, &taskp );

	CKERR( info, "pvm_tasks", NOHOST, return( FAILED ) );

	orig_ntask = ntask;

#ifdef DEBUG
	for ( i=0 ; i < ntask ; i++ )
	{
		printf( "Task: %s Tid: 0x%x\n",
				taskp[i].ti_a_out, taskp[i].ti_tid );
	}
#endif

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	if ( (info = spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskDefault, NOHOST, length, tids ))
			< 0 )
	{
		result = FAILED;
		goto fail;
	}

	for( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( -1, ACKNOWLEDGE, -1, VERBOSE );

		if ( ack < 0 )
		{
			errprintf( "\tpvm_recv(): Failed to receive all acks\n" );
	 		result = FAILED;
	 		goto fail;
		}

		if ( master_recv( -1, ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
			result = FAILED;
	}

	info = pvm_tasks( 0, &ntask, &taskp );

	CKERR( info, "pvm_tasks", NOHOST, { result = FAILED; goto fail; } );

	if ( ntask != length + orig_ntask )
	{
		errprintf( "\tpvm_task(): Actual # of tasks %d != %d\n",
				ntask, length + orig_ntask );

		result = FAILED;
	}

	for ( i=0 ; i < length ; i++ )
	{
		info = pvm_tasks( tids[i], &ntask, &taskp );

		CKERR( info, "pvm_tasks", hostpool[ array_of_index[i] ],
				{ result = FAILED; continue; } );

		if ( ntask != 1 )
		{
			errprintf( "\ttpvm_tasks():  single task -> ntask=%d!\n",
					ntask );
			result = FAILED;
		}

		else if ( taskp[0].ti_tid != tids[i]
				|| taskp[0].ti_host != pvm_tidtohost( tids[i] ) )
		{
	 		errprintf( "\tpvm_tasks():  incorrect task information " );
			errprintf( "tid(0x%x/0x%x) host(0x%x/0x%x)\n",
					taskp[0].ti_tid, tids[i],
					taskp[0].ti_host, pvm_tidtohost( tids[i] ) );
	 		result = FAILED;
		}

		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

fail:

	free( tids );

	return( result );
}


/**************/
/*  pvm_exit  */
/**************/

int
test_exit( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		info;
	int		tid;
	int		result = PASSED;

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;

		info = timeout_wait_kill( tid, array_of_index[i] );

		if ( info < 0 )
		{
			errprintf(
				"\tpvm_exit(): slave process did not exited in time " );
			errprintf( "on host %s\n",
				hostpool[ array_of_index[i] ] );
			result = FAILED;
			pvm_kill( tid );
		}

		/* child will be killed anyway with the deletion of the host */
	}

	return( result );
}


/***************************/
/*  several group started  */
/***************************/

int
test_multi_pvmgs( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct pvmtaskinfo *taskp;

	int		*tids;

	int		result = PASSED;
	int		num_msg = 0;
	int		ntask;
	int		info;
	int		ack;
	int		i;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
			result = FAILED;

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;
	}

	if ( result == FAILED )
		goto fail;

	info = pvm_tasks( 0, &ntask, &taskp );

	CKERR( info, "pvm_tasks", NOHOST, goto fail  );

	for ( i=0 ; i < ntask ; i++ )
		if ( strcmp( taskp[i].ti_a_out, "pvmgs" ) == 0 )
			num_msg++;

	if ( num_msg == 0 )
	{
		errprintf( "No group server started\n" );
		result = FAILED;
	}

	if ( num_msg > 1 )
	{
		errprintf( "%d group servers started!\n", num_msg );
		result = FAILED;
	}

fail:

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/*****************************/
/*  pvm_joingroup incorrect  */
/*****************************/

int
test_joingroup_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		result = PASSED;
	int		info;

	info = pvm_joingroup( NULL );

	EXPECTED( "pvm_joingroup", PvmNullGroup, info, NOHOST,
			result = FAILED );

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );

	CKERR( info, "pvm_joingroup", NOHOST, return( FAILED ) );

	info = pvm_joingroup( group );

	EXPECTED( "pvm_joingroup", PvmDupGroup, info, NOHOST,
			return( FAILED ) );

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST,
			{
				errprintf( "*** May affect future behavior ***\n" );
				return( FAILED );
			} );

	return( result );
}


/***************************/
/*  pvm_lvgroup incorrect  */
/***************************/

int
test_lvgroup_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		info;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );

	CKERR( info, "pvm_joingroup", NOHOST, return( FAILED ) );

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST, return( FAILED ) );

	info = pvm_lvgroup( group );

	EXPECTED( "pvm_lvgroup", PvmNoGroup, info, NOHOST,
			return( FAILED ) );

	return( PASSED );
}


/*************************************************/
/*  pvm_joingroup()  pvm_lvgroup()  pvm_gsize()  */
/*************************************************/

int
test_gsize( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		info;
	int		ack;
	int		i;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", PvmNoGroup, info, NOHOST, result = FAILED );

	info = pvm_joingroup( group );

	CKERR( info, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", 1, info, NOHOST,
			{ result = FAILED; goto fail; } );

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;
	}

	if ( result == FAILED )
		goto fail;

	info = pvm_gsize( group );

	CKERR( info, "pvm_gsize", NOHOST,
			{ result = FAILED; goto fail; } );

	EXPECTED( "pvm_gsize", length + 1, info, NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );

		ack = timeout_wait_kill( tids[i], -1 );

		if ( ack < 0 )
			result = FAILED;
	}

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", 1, info, NOHOST,
			{ result = FAILED; goto fail; } );

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 0;

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", PvmNoGroup, info, NOHOST, result = FAILED );

fail:

	if ( ingroup )
		pvm_lvgroup( group );

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/**********************/
/*  group completion  */
/**********************/

int
test_group_cleanup( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		info;
	int		ack;
	int		i;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", PvmNoGroup, info, NOHOST, result = FAILED );

	info = pvm_joingroup( group );

	CKERR( info, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", 1, info, NOHOST,
			{ result = FAILED; goto fail; } );

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;
	}

	if ( result == FAILED )
		goto fail;

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_kill( tids[i], array_of_index[i] );

		if ( ack < 0 )
			result = FAILED;
	}

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 0;

	info = pvm_gsize( group );

	EXPECTED( "pvm_gsize", PvmNoGroup, info, NOHOST, result = FAILED );

fail:

	if ( ingroup )
		pvm_lvgroup( group );
	
	free( tids );

	return( result );
}


/********************************/
/*  pvm_getinst() pvm_gettid()  */
/********************************/

int
test_getinst_gettid( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*tids;

	int		result = PASSED;
	int		proc_inst;
	int		proc_tid;
	int		mytid;
	int		info;
	int		ack;
	int		i;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_gettid( group, 0 );

	EXPECTED( "pvm_gettid", PvmNoGroup, info, NOHOST, result = FAILED );

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	mytid = pvm_mytid();

	CKERR( mytid, "pvm_mytid", NOHOST,
			{ result = FAILED; goto fail; } );

	info = pvm_getinst( group, mytid );

	EXPECTED( "pvm_getinst", PvmNotInGroup, info, NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		proc_inst = pvm_getinst( group, tids[i] );

		CKERR( proc_inst, "pvm_getinst", hostpool[ array_of_index[i] ],
				{ result = FAILED; goto fail; } );

		proc_tid = pvm_gettid( group, proc_inst );

		EXPECTED( "pvm_gettid", tids[i], proc_tid,
				hostpool[ array_of_index[i] ],
				{ result = FAILED; goto fail; } );
	}

	info = pvm_gettid( group, length );

	EXPECTED( "pvm_gettid", PvmNoInst, info, NOHOST, result = FAILED );

	for ( i=0 ; i < length ; i++ )
	{
		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );

		ack = timeout_wait_kill( tids[i], -1 );

		if ( ack < 0 )
			result = FAILED;
	}

	info = pvm_gettid( group, 0 );

	EXPECTED( "pvm_gettid", PvmNoGroup, info, NOHOST, result = FAILED );

fail:

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/************************************************/
/*  pvm_initsend() pvm_getsbuf() pvm_freebuf()  */
/************************************************/

int
test_buffer( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		bufid;
	int		bufid2;
	int		info;

	bufid = pvm_initsend( PvmDataDefault );

	CKERR( bufid, "pvm_initsend", NOHOST, return( FAILED ) );

	bufid2 = pvm_getsbuf();

	EXPECTED( "pvm_getsbuf", bufid, bufid2, NOHOST, return( FAILED ) );

	info = pvm_freebuf( bufid );

	CKERR( info, "pvm_freebuf", NOHOST, return( FAILED ) );

	info = pvm_freebuf( bufid );

	EXPECTED( "pvm_freebuf", PvmNoSuchBuf, info, NOHOST,
			return( FAILED ) );

	return( PASSED );
}


/****************************/
/*  pvm_pkxx() pvm_upkxx()  */
/****************************/

int
test_packing( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	void	*data_in;
	void	*data_out;

	char	*char_in;
	char	*char_out;
	char	*byte_in;
	char	*byte_out;
	int		*int_in;
	int		*int_out;
	short	*short_in;
	short	*short_out;
	long	*long_in;
	long	*long_out;
	u_int	*uint_in;
	u_int	*uint_out;
	u_short	*ushort_in;
	u_short	*ushort_out;
	u_long	*ulong_in;
	u_long	*ulong_out;
	float	*float_in;
	float	*float_out;
	double	*double_in;
	double	*double_out;
	float	*comp_in;
	float	*comp_out;
	double	*dcomp_in;
	double	*dcomp_out;

	int		ack;
	int		data_format;
	int		info, info1, info2;
	int		nelems;
	int		result = PASSED;
	int		send_mode;
	int		recv_mode;
	int		tid;
	int		i;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	nelems		=	args[2];
	data_format	=	args[3];

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	data_in = (void *) malloc( ( (unsigned) ( nelems * 2 ) )
			* sizeof( double ) );
	if ( !memcheck( data_in, "Packing Data" ) )
		return( FAILED );

	data_out = (void *) malloc( ( (unsigned) ( nelems * 2 ) )
			* sizeof( double ) );
	if ( !memcheck( data_out, "Unpacking Data" ) )
	{
		free( data_in );
		return( FAILED );
	}

	/* Start Slave */

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskHost, hostpool[ array_of_index[ length - 1 ] ],
			1, &tid )
			< 0 )
	{
		free( data_out );
		free( data_in );

		return( FAILED );
	}

	/* Acknoledgement from slave */

	ack = timeout_wait_message( tid, ACKNOWLEDGE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* CHAR */

	if ( send_mode != PSEND && recv_mode != PRECV )
	{
		char_out = (char *) data_out;
		char_in = (char *) data_in;

		for ( i=0 ; i < nelems - 1 ; i++ )
			char_out[i] = 'a' + ( i % 26 );

		char_out[ nelems - 1 ] = '\0';

		if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
				(void *) char_out, nelems, PVM_STR, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		ack = timeout_wait_message( tid, TEST_VALUE,
				array_of_index[ length - 1 ], VERBOSE );

		if ( ack < 0 )
		{
			if ( vflag ) errprintf( "\t -- CHAR --\n" );
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tid, TEST_VALUE, recv_mode,
				(void *) char_in, nelems, PVM_STR, (int *) NULL,
				hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( strcmp( char_out, char_in ) != 0 )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkstr() and pvm_upkstr() " );
			errprintf( "(%s != %s)\n", char_in, char_out );
			result = FAILED;
		}
	}

	/* BYTE */

	byte_out = (char *) data_out;
	byte_in = (char *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		byte_out[i] = 'A' + ( i % 26 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) byte_out, nelems, PVM_BYTE, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- BYTE --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) byte_in, nelems, PVM_BYTE, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( byte_out[i] != byte_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkbyte() and pvm_upkbyte() " );
			errprintf( "(%c != %c)\n", byte_in[i], byte_out[i] );
			result = FAILED;
		}
	}

	/* INT */

	int_out = (int *) data_out;
	int_in = (int *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		int_out[i] = i + 1;

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) int_out, nelems, PVM_INT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- INT --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) int_in, nelems, PVM_INT, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( int_in[i] != int_out[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkint() and pvm_upkint() " );
			errprintf( "(%d != %d)\n", int_in[i], int_out[i] );
			result = FAILED;
		}
	}

	/* SHORT */

	short_out = (short *) data_out;
	short_in = (short *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		short_out[i] = (short) ( i + 1 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) short_out, nelems, PVM_SHORT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- SHORT --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) short_in, nelems, PVM_SHORT, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( short_out[i] != short_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkshort() and pvm_upkshort() " );
			errprintf( "(%d != %d)\n", short_in[i], short_out[i] );
			result = FAILED;
		}
	}

	/* LONG */

	long_out = (long *) data_out;
	long_in = (long *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		long_out[i] = (long) ( i + 1 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) long_out, nelems, PVM_LONG, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- LONG --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) long_in, nelems, PVM_LONG, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( long_out[i] != long_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pklong() and pvm_upklong() " );
			errprintf( "(%ld != %ld)\n", long_in[i], long_out[i] );
			result = FAILED;
		}
	}

	/* U_INT */

	uint_out = (unsigned int *) data_out;
	uint_in = (unsigned int *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		uint_out[i] = (unsigned int) ( i + 1 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) uint_out, nelems, PVM_UINT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- U_INT --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) uint_in, nelems, PVM_UINT, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( uint_out[i] != uint_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkuint() and pvm_upkuint() " );
			errprintf( "(%u != %u)\n", uint_in[i], uint_out[i] );
			result = FAILED;
		}
	}

	/* U_SHORT */

	ushort_out = (unsigned short *) data_out;
	ushort_in = (unsigned short *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		ushort_out[i] = (unsigned short) ( i + 1 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) ushort_out, nelems, PVM_USHORT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- U_SHORT --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) ushort_in, nelems, PVM_USHORT, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( ushort_out[i] != ushort_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkushort() and pvm_upkushort() " );
			errprintf( "(%u != %u)\n", ushort_in[i], ushort_out[i] );
			result = FAILED;
		}
	}

	/* U_LONG */

	ulong_out = (unsigned long *) data_out;
	ulong_in = (unsigned long *) data_in;

	for ( i=0 ; i < nelems ; i++ )
		ulong_out[i] = (unsigned long) ( i + 1 );

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) ulong_out, nelems, PVM_ULONG, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- U_LONG --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) ulong_in, nelems, PVM_ULONG, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( ulong_out[i] != ulong_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkulong() and pvm_upkulong() " );
			errprintf( "(%lu != %lu)\n", ulong_in[i], ulong_out[i] );
			result = FAILED;
		}
	}

	/* FLOAT */

	float_out = (float *) data_out;
	float_in = (float *) data_in;

	for ( i=0 ; i < nelems ; i++ )
	{
		float_out[i] = ((float) ( i + 1 ))
				+ ((float) i / (float) 10000.0);
	}

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) float_out, nelems, PVM_FLOAT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- FLOAT --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) float_in, nelems, PVM_FLOAT, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( float_out[i] != float_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkfloat() and pvm_upkfloat() " );
			errprintf( "(%f != %f)\n", float_in[i], float_out[i] );
			result = FAILED;
		}
	}

	/* DOUBLE */

	double_out = (double *) data_out;
	double_in = (double *) data_in;

	for ( i=0 ; i < nelems ; i++ )
	{
		double_out[i] = ((double) ( i + 1 ))
				+ ((double) i / (double) 10000.0);
	}

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) double_out, nelems, PVM_DOUBLE, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- DOUBLE --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) double_in, nelems, PVM_DOUBLE, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i++ )
	{
		if ( double_out[i] != double_in[i] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkdouble() and pvm_upkdouble() " );
			errprintf( "(%lf != %lf)\n", double_in[i], double_out[i] );
			result = FAILED;
		}
	}

	/* CPLX */

	comp_out = (float *) data_out;
	comp_in = (float *) data_in;

	for ( i=0 ; i < nelems ; i += 2 )
	{
		comp_out[i] = ((float) ( i + 1 ))
				+ ((float) i / (float) 10000.0);
		comp_out[i+1] = ((float) -1.0) * ( ((float) ( i + 1 ))
				+ ((float) i / (float) 10000.0) );
	}

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) comp_out, ( nelems == 1 ) ? nelems : nelems / 2,
			PVM_CPLX, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- CPLX --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) comp_in, ( nelems == 1 ) ? nelems : nelems / 2,
			PVM_CPLX, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i += 2 )
	{
		if ( comp_out[i] != comp_in[i]
				|| comp_out[i+1] != comp_in[i+1] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkcplx() and pvm_upkcplx() " );
			errprintf( "(%f/%f != %f/%f)\n",
					comp_in[i], comp_in[1],
					comp_out[i+i], comp_out[i+1] );
			result = FAILED;
		}
	}

	/* DCPLX */

	dcomp_out = (double *) data_out;
	dcomp_in = (double *) data_in;

	for ( i=0 ; i < nelems ; i += 2 )
	{
		dcomp_out[i] = ((double) ( i + 1 ))
				+ ((double) i / (double) 10000.0);
		dcomp_out[i+1] = ((double) -1.0) * ( ((double) ( i + 1 ))
				+ ((double) i / (double) 10000.0) );
	}

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, data_format,
			(void *) dcomp_out, ( nelems == 1 ) ? nelems : nelems / 2,
			PVM_DCPLX, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		if ( vflag ) errprintf( "\t -- DCPLX --\n" );
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) dcomp_in, ( nelems == 1 ) ? nelems : nelems / 2,
			PVM_DCPLX, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < nelems ; i += 2 )
	{
		if ( dcomp_out[i] != dcomp_in[i]
				|| dcomp_out[i+1] != dcomp_in[i+1] )
		{
			errprintf( "Data error between " );
			errprintf( "pvm_pkdcplx() and pvm_upkdcplx() " );
			errprintf( "(%lf/%lf != %lf/%lf)\n",
					dcomp_in[i], dcomp_in[1],
					dcomp_out[i+i], dcomp_out[i+1] );
			result = FAILED;
		}
	}

fail:

	pvm_kill( tid );

	free( data_out );
	free( data_in );

	return( result );
}


/*******************/
/*  pvm_getrbuf()  */
/*******************/

int
test_getrbuf( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		*tids;
	int		info;
	int		result = PASSED;
	int		bufid;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( (info = master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] )) < 0 )
		{
			result = FAILED;
			continue;
		}

		bufid = pvm_getrbuf();

		EXPECTED( "pvm_getrbuf", info, bufid,
				hostpool[ array_of_index[i] ], result = FAILED );

		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	free( tids );

	return( result );
}


/*******************/
/*  pvm_bufinfo()  */
/*******************/

int
test_bufinfo( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		*tids;
	int		info;
	int		result = PASSED;
	int		bufid;
	int		proc_tid;
	int		bytes;
	int		msgtag;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			pvm_kill( tids[i] );
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( -1, ACKNOWLEDGE, array_of_index[i],
				VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tids[i] );
			continue;
		}

		if ( (bufid = master_recv( -1, ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] )) < 0 )
		{
			result = FAILED;
			pvm_kill( tids[i] );
			continue;
		}

		info = pvm_bufinfo( bufid, &bytes, &msgtag, &proc_tid );

		CKERR( info, "pvm_bufinfo", hostpool[ array_of_index[i] ],
				{ result = FAILED; pvm_kill( tids[i] ); continue; } );

		if ( bytes != 4 )
		{
			errprintf( "\tpvm_bufinfo():  Erroneous message length " );
			errprintf( "(%d != 4) for host %s\n",
					bytes, hostpool[ array_of_index[i] ] );
			pvm_kill( tids[i] );
			result = FAILED;
		}

		if ( msgtag != ACKNOWLEDGE )
		{
			errprintf( "\tpvm_bufinfo():  Erroneous message tag " );
			errprintf( "(%d != %d) for host %s\n",
					msgtag, ACKNOWLEDGE,
					hostpool[ array_of_index[i] ] );
			pvm_kill( tids[i] );
			result = FAILED;
		}

		if ( proc_tid != tids[i] )
		{
			errprintf( "\tpvm_bufinfo():  Erroneous task id " );
			errprintf( "(t%x != t%x) for host %s\n",
					proc_tid, tids[i], hostpool[ array_of_index[i] ] );
			pvm_kill( tids[i] );
			result = FAILED;
		}

		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	free( tids );

	return( result );
}


/*****************/
/*  pvm_mytid()  */
/*****************/

int
test_mytid( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		*tids;
	int		info;
	int		result = PASSED;
	int		proc_tid;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( -1, ACKNOWLEDGE, array_of_index[i],
				VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tids[i] );
			continue;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) &proc_tid, 1, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			pvm_kill( tids[i] );
			continue;
		}

		EXPECTED( "pvm_upkint", tids[i], proc_tid,
				hostpool[ array_of_index[i] ], result = FAILED );

		info = pvm_kill( tids[i] );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	free( tids );

	return( result );
}


/*****************/
/*  pvm_mkbuf()  */
/*****************/

int
test_mkbuf( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		bufid;
	int		info;
	int		dummy=1;

	bufid = pvm_mkbuf( PvmDataDefault );

	CKERR( bufid, "pvm_mkbuf", NOHOST, return( FAILED ) );

	info = pvm_setsbuf( bufid );

	CKERR( info, "pvm_setsbuf", NOHOST, return( FAILED ) );

	info = pvm_pkint( &dummy, 1, 1 );

	CKERR( info, "pvm_pkint", NOHOST, return( FAILED ) );

	info = pvm_freebuf( bufid );

	CKERR( info, "pvm_freebuf", NOHOST, return( FAILED ) );

	return( PASSED );
}


/*************************/
/*  send/recv coherency  */
/*************************/

int
test_coherency( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct pvmhostinfo *hostp;

	double res_dummy[10];
	double dummy[10];

	int		amsg[3];
	int		ack;
	int		tid;
	int		info;
	int		narch;
	int		nhost;
	int		result = PASSED;
	int		send_mode;
	int		recv_mode;
	int		i;

	send_mode = args[0];
	recv_mode = args[1];

	if ( send_mode == PSEND || recv_mode == PRECV )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	/* Initialization */

	for ( i=0 ; i < 10 ; i++ )
		dummy[i] = ( 1.0 + (double) i ) / 10.0;

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskHost, hostpool[ array_of_index[ length - 1 ] ],
			1, &tid )
			< 0 )
	{
		return( FAILED );
	}

	/* Acknoledgement from slave */

	ack = timeout_wait_message( tid, ACKNOWLEDGE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* Just Do It */

	if ( master_send( &(tid), 1, TEST_VALUE, send_mode, PvmDataDefault,
			(void *) dummy, 10, PVM_DOUBLE,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	ack = timeout_wait_message( tid, TEST_VALUE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, TEST_VALUE, recv_mode,
			(void *) res_dummy, 10, PVM_DOUBLE, amsg,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( i=0 ; i < 10 ; i++ )
	{
		if ( res_dummy[i] != dummy[i] )
		{
			errprintf(
					"\tMismatch:  Value #%d (%lf != %lf) for host %s\n",
					i, res_dummy[i], dummy[i],
					hostpool[ array_of_index[ length - 1 ] ] );
			result = FAILED;
		}
	}

	if ( recv_mode == PRECV )
	{
		if ( amsg[0] != tid )
		{
			errprintf( "\tpvm_precv():  incorrect source tid " );
			errprintf( "0x%x != 0x%x\n", amsg[0], tid );
			result = FAILED;
		}

		if ( amsg[1] != TEST_VALUE )
		{
			errprintf( "\tpvm_precv():  incorrect message tag " );
			errprintf( "%d != %d\n", amsg[1], TEST_VALUE );
			result = FAILED;
		}

		if ( amsg[2] != 10 * sizeof( double ) )
		{
			errprintf( "\tpvm_precv():  incorrect message length " );
			errprintf( "%d != %d\n", amsg[2], 10 * sizeof( double ) );
			result = FAILED;
		}
	}

fail:

	pvm_kill( tid );

	return( result );
}


/*******************/
/*  pvm_setsbuf()  */
/*******************/

int
test_setsbuf( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		tid;
	int		info;
	int		result = PASSED;
	int		bufid;
	int		dummy=12;
	int		res_dummy;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		bufid = pvm_mkbuf( PvmDataDefault );

		CKERR( bufid, "pvm_mkbuf", NOHOST,
				{ result = FAILED; pvm_kill( tid ); continue; } );

		info = pvm_setsbuf( bufid );

		CKERR( info, "pvm_setsbuf", NOHOST,
				{
					pvm_freebuf( bufid );
					result = FAILED;
					pvm_kill( tid );
					continue;
				} );

		info = pvm_pkint( &dummy, 1, 1 );

		CKERR( info, "pvm_pkint", NOHOST,
				{
					pvm_freebuf( bufid );
					result = FAILED;
					pvm_kill( tid );
					continue;
				} );

		info = pvm_send( tid, TEST_VALUE );

		CKERR( info, "pvm_send", NOHOST,
				{
					pvm_freebuf( bufid );
					result = FAILED;
					pvm_kill( tid );
					continue;
				} );

		ack = timeout_wait_message( tid, TEST_VALUE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_freebuf( bufid );
			pvm_kill( tid );
			continue;
		}

		if ( master_recv( tid, TEST_VALUE, RECV,
				(void *) &res_dummy, 1, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			pvm_freebuf( bufid );
			result = FAILED;
			pvm_kill( tid );
			continue;
		}

		if ( res_dummy != dummy )
		{
			errprintf( "Value Mismatch (%d != %d) for host %s\n",
					res_dummy, dummy, hostpool[ array_of_index[i] ] );
			result = FAILED;
		}

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/*******************/
/*  pvm_setrbuf()  */
/*******************/

int
test_setrbuf( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		tid;
	int		info;
	int		result = PASSED;
	int		bufid;
	int		res_dummy;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tid, ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tid, ACKNOWLEDGE, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			continue;
		}

		bufid = pvm_mkbuf( PvmDataDefault );

		CKERR( bufid, "pvm_mkbuf", NOHOST,
				{ result = FAILED; continue; } );

		info = pvm_setrbuf( bufid );

		CKERR( info, "pvm_setrbuf", NOHOST,
			{ pvm_freebuf( bufid ); result = FAILED; continue; } );

		ack = timeout_wait_message( tid, TEST_VALUE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			pvm_freebuf( bufid );
			continue;
		}

		if ( master_recv( tid, TEST_VALUE, RECV,
				(void *) &res_dummy, 1, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			pvm_freebuf( bufid );
			result = FAILED;
			continue;
		}

		if ( res_dummy != tid )
		{
			errprintf( "Value Mismatch (%d != %d) for host %s\n",
					res_dummy, tid, hostpool[ array_of_index[i] ] );
			result = FAILED;
		}

		info = pvm_kill( tid );

		CKERR( info, "pvm_kill", NOHOST, result = FAILED );
	}

	return( result );
}


/*****************************/
/*  pvm_set?buf() incorrect  */
/*****************************/

int
test_setbuf_incorrect( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		bufid;
	int		info;
	int		result = PASSED;

	bufid = pvm_mkbuf( PvmDataDefault );

	CKERR( bufid, "pvm_mkbuf", NOHOST, result = FAILED );

	info = pvm_freebuf( bufid );

	CKERR( info, "pvm_freebuf", NOHOST, result = FAILED );

	info = pvm_setsbuf( bufid );

	EXPECTED( "pvm_setsbuf", PvmNoSuchBuf, info, NOHOST,
			result = FAILED );

	info = pvm_setrbuf( bufid );

	EXPECTED( "pvm_setrbuf", PvmNoSuchBuf, info, NOHOST,
			result = FAILED );

	return( result );
}


/*****************/
/*  pvm_bcast()  */
/*****************/

int
test_bcast( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*tids;

	int		result = PASSED;
	int		info;
	int		ack;
	int		i;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_bcast( group, TEST_CAST );

	EXPECTED( "pvm_bcast", PvmNoGroup, info, NOHOST, return( FAILED ) );

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_initsend( PvmDataDefault );

	CKERR( info, "pvm_initsend", NOHOST,
			{ result = FAILED; goto fail; } );

	info = pvm_bcast( group, TEST_CAST );

	CKERR( info, "pvm_bcast", NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		ack= timeout_wait_message( tids[i], TEST_CAST,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], TEST_CAST, RECV, (void *) NULL, 0, 0,
				(int *) NULL, hostpool[ array_of_index[i] ] ) < 0 )
			result = FAILED;
	}

fail:

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/*******************/
/*  pvm_barrier()  */
/*******************/

int
test_barrier( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		info;
	int		ack;
	int		i;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );

	CKERR( info, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_send( &(tids[i]), 1, TEST_BARRIER,
				SEND, PvmDataDefault,
				(void *) &length, 1, PVM_INT, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_barrier( group, length + 1 );

	CKERR( info, "pvm_barrier", NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( tids[i], TEST_BARRIER,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], TEST_BARRIER, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		ack = timeout_wait_kill( tids[i], array_of_index[i] );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST, result = FAILED );

	ingroup = 0;

fail:

	if ( ingroup )
		pvm_lvgroup( group );

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/******************/
/*  pvm_reduce()  */
/******************/

int
test_reduce( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*data;
	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		count = 10;
	int		info;
	int		root;
	int		ack;
	int		i;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	data = (int *) malloc( count * sizeof( int ) );
	if ( !memcheck( data, "Reduce Data Array" ) )
	{
		free( tids );
		return( FAILED );
	}

	for ( i=0 ; i < count ; i++ )
		data[i] = i;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_reduce( PvmSum, data, count, PVM_INT, TEST_REDUCE,
			group, 0 );

	EXPECTED( "pvm_reduce", PvmNoGroup, info, NOHOST,
			{ result = FAILED; goto fail; } );

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
			< 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	/* Acknoledgement from slave */

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
	 		result = FAILED;
	 		goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_reduce( PvmSum, data, count, PVM_INT, TEST_REDUCE,
			group, 0 );

	EXPECTED( "pvm_reduce", PvmNotInGroup, info, NOHOST,
			{ result = FAILED; goto fail; } );

	root = pvm_joingroup( group );

	CKERR( root, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	for ( i=0 ; i < length ; i++ )
	{
		if ( master_send( &(tids[i]), 1, TEST_REDUCE,
				SEND, PvmDataDefault,
				(void *) &root, 1, PVM_INT, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_reduce( PvmSum, data, count, PVM_INT, TEST_REDUCE,
			group, root );

	CKERR( info, "pvm_reduce", NOHOST,
			{ result = FAILED; goto fail; } );

	/* Verify results */

	for ( i=0 ; i < count ; i++ )
	{
		if ( data[i] != i * ( length + 1 ) )
		{
			errprintf( "\tpvm_reduce():  Incorrect result " );
			errprintf( "at position %d (%d != %d)\n",
					i, data[i], i * ( length + 1 ) );
			result = FAILED;
		}
	}

fail:

	if ( ingroup )
		pvm_lvgroup( group );

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );
	free( data );

	return( result );
}


/******************/
/*  pvm_gather()  */
/******************/

int
test_gather( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*results;
	int		*data;
	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		count = 10;
	int		info;
	int		root;
	int		i, j;
	int		ack;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	data = (int *) malloc( count * sizeof( int ) );
	if ( !memcheck( data, "Gather Data Array" ) )
	{
		free( tids );
		return( FAILED );
	}

	results = (int *) malloc( count * ( length + 1 ) * sizeof( int ) );
	if ( !memcheck( results, "Gather Results Array" ) )
	{
		free( tids );
		free( data );
		return( FAILED );
	}

	for ( i=0 ; i < count ; i++ )
		data[i] = i;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_gather( results, data, count, PVM_INT, TEST_GATHER,
			group, 0 );

	EXPECTED( "pvm_gather", PvmNoGroup, info, NOHOST, result = FAILED );

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_gather( results, data, count, PVM_INT, TEST_GATHER,
			group, 0 );

	EXPECTED( "pvm_gather", PvmNotInGroup, info, NOHOST,
			result = FAILED );

	root = pvm_joingroup( group );

	CKERR( root, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	for ( i=0 ; i < length ; i++ )
	{
		if ( master_send( &(tids[i]), 1, TEST_GATHER,
				SEND, PvmDataDefault,
				(void *) &root, 1, PVM_INT, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_gather( results, data, count, PVM_INT, TEST_GATHER,
			group, root );

	CKERR( info, "pvm_gather", NOHOST,
			{ result = FAILED; goto fail; } );

	/* Check result */

	for ( i=0 ; i < length + 1 ; i++ )
	{
		for ( j=0 ; j < count ; j++ )
		{
			if ( results[i * count + j] != j )
			{
				errprintf( "\tpvm_gather(): Incoherent data " );
				errprintf( "(%d != %d )\n", results[i * count + j], j );
				result = FAILED;
			}
		}
	}

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST, result = FAILED );

	ingroup = 0;

fail:

	if ( ingroup )
		pvm_lvgroup( group );

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );
	free( data );
	free( results );

	return( result );
}


/*******************/
/*  pvm_scatter()  */
/*******************/

int
test_scatter( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	char	group[255];

	int		*results;
	int		*data;
	int		*tids;

	int		result = PASSED;
	int		ingroup = 0;
	int		count = 5;
	int		info;
	int		root;
	int		i, j;
	int		ack;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	data = (int *) malloc( count * ( length + 1 ) * sizeof( int ) );
	if ( !memcheck( data, "Scatter Data Array" ) )
	{
		free( tids );
		return( FAILED );
	}

	results = (int *) malloc( count * sizeof( int ) );
	if ( !memcheck( results, "Scatter Results Array" ) )
	{
		free( tids );
		free( data );
		return( FAILED );
	}

	for ( i=0 ; i < length + 1 ; i++ )
		for ( j=0 ; j < count ; j++ )
			data[ ( i * count ) + j ] = j;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_scatter( results, data, count, PVM_INT, TEST_SCATTER,
			group, 0 );

	EXPECTED( "pvm_scatter", PvmNoGroup, info, NOHOST,
			result = FAILED );

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_scatter( results, data, count, PVM_INT, TEST_SCATTER,
			group, 0 );

	EXPECTED( "pvm_scatter", PvmNotInGroup, info, NOHOST,
			result = FAILED );

	root = pvm_joingroup( group );

	CKERR( root, "pvm_joingroup", NOHOST,
			{ result = FAILED; goto fail; } );

	ingroup = 1;

	for ( i=0 ; i < length ; i++ )
	{
		if ( master_send( &(tids[i]), 1, TEST_SCATTER,
				SEND, PvmDataDefault,
				(void *) &root, 1, PVM_INT, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}


	info = pvm_scatter( results, data, count, PVM_INT, TEST_SCATTER,
			group, root );

	CKERR( info, "pvm_scatter", NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		if ( master_send( &(tids[i]), 1, TEST_SCATTER,
				SEND, PvmDataDefault,
				(void *) NULL, 0, 0,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	/* Check result */

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( tids[i], TEST_SCATTER,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], TEST_SCATTER, RECV,
				(void *) results, count, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			continue;
		}

		for ( j=0 ; j < count ; j++ )
		{
			if ( data[j] != j )
			{
				errprintf(
					"\tpvm_scatter():  Erroneous value scattered " );
				errprintf( "(%d != %d)\n", data[j], j );
				result = FAILED;
			}
		}
	}

	info = pvm_lvgroup( group );

	CKERR( info, "pvm_lvgroup", NOHOST, result = FAILED );

	ingroup = 0;

fail:

	if ( ingroup )
		pvm_lvgroup( group );

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );
	free( data );
	free( results );

	return( result );
}


/*****************/
/*  pvm_mcast()  */
/*****************/

int
test_mcast( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		ack;
	int		*tids;
	int		info;
	int		result = PASSED;

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			goto fail;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			goto fail;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_initsend( PvmDataDefault );

	CKERR( info, "pvm_initsend", NOHOST,
			{ result = FAILED; goto fail; } );

	info = pvm_mcast( tids, length, TEST_MCAST );

	CKERR( info, "pvm_mcast", NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( tids[i], TEST_MCAST,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], TEST_MCAST, RECV,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
		}
	}

fail:

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/*****************/
/*  pvm_trecv()  */
/*****************/

int
test_trecv( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i;
	int		tid;
	int		bufid;
	int		result = PASSED;

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		bufid = pvm_trecv( tid, ACKNOWLEDGE, &TRECV_TM );

		CKERR( bufid, "pvm_trecv", hostpool[ array_of_index[i] ],
				{ result = FAILED; continue; } );

		EXPECTED( "pvm_trecv", 0, bufid, NOHOST, result = FAILED );

		if ( master_send( &tid, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
				(void *) NULL, 0, 0, NOHOST ) < 0 )
		{
			result = FAILED;
			continue;
		}

		bufid = pvm_trecv( tid, ACKNOWLEDGE, &TRECV_TM );

		CKERR( bufid, "pvm_trecv", hostpool[ array_of_index[i] ],
				{ result = FAILED; continue; } );

		if ( bufid == 0 )
		{
			errprintf(
				"\tpvm_trecv(): Didn't receive message in time?\n" );
		}
	}

	return( result );
}


/*****************/
/*  pvm_nrecv()  */
/*****************/

int
test_nrecv( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		i, j;
	int		tid;
	int		bufid;
	int		result = PASSED;

	for ( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ], 1, &tid )
				< 0 )
		{
			result = FAILED;
			continue;
		}

		/* Acknoledgement from slave */

		bufid = pvm_nrecv( tid, ACKNOWLEDGE );

		CKERR( bufid, "pvm_nrecv", hostpool[ array_of_index[i] ],
				{ result = FAILED; continue; } );

		EXPECTED( "pvm_nrecv", 0, bufid, NOHOST, result = FAILED );

		if ( master_send( &tid, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
				(void *) NULL, 0, 0, NOHOST ) < 0 )
		{
			result = FAILED;
			continue;
		}

		for ( j=0 ; j < SEC_NRECV ; j++ )
		{
			bufid = pvm_nrecv( tid, ACKNOWLEDGE );

			CKERR( bufid, "pvm_nrecv", hostpool[ array_of_index[i] ],
					{ result = FAILED; continue; } );

			if ( bufid > 0 )
				break;

			sleep( 1 );
		}

		if ( j == SEC_NRECV )
		{
			errprintf(
				"\tpvm_nrecv(): Didn't receive message in time?\n" );
		}
	}

	return( result );
}


/*************************/
/*  Latency / Bandwidth  */
/*************************/

int
test_perf( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct timeval t1, t2;

	double Average_band;
	double Average_time;
	double Total_size;
	double Total_time;

	int		outbuf[MAX_MESSAGE_LENGTH];
	int		inbuf[MAX_MESSAGE_LENGTH];

	int		start, end, step;
	int		result = PASSED;
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		data[3];
	int		count;
	int		info;
	int		ack;
	int		tag;
	int		tid;
	int		i, j;
	int		b;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	data[0] = max_size;
	data[1] = start_size;
	data[2] = incr;

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	if ( vflag )
	{
		errprintf( "\n%s( %s ) < -- > %s( %s )\n",
				hostpool[0], hostarch[0],
				hostpool[ array_of_index[ length - 1 ] ],
				hostarch[ array_of_index[ length - 1 ] ] );
	}

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskHost, hostpool[ array_of_index[ length - 1 ] ],
			1, &tid )
			< 0 )
	{
		return( FAILED );
	}

	/* Acknoledgement from slave */

	ack = timeout_wait_message( tid, ACKNOWLEDGE,
			array_of_index[ length - 1 ], VERBOSE );

	if ( ack < 0 )
	{
		result = FAILED;
		goto fail;
	}

	if ( master_recv( tid, ACKNOWLEDGE, recv_mode,
			(void *) NULL, 0, 0, (int *) NULL,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* Send Range */

	if ( master_send( &tid, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) data, 3, PVM_INT,
			hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* Just Do It */

	for ( b=0 ; b < 3 ; b++ )
	{
		if ( !b )
		{
			start = 0;
			end = 1;
			step = 1;

			tag = LATENCY;
		}

		else
		{
			if ( b == 1 )
			{
				if ( vflag ) errprintf( "Daemon routing\n" );

				info = pvm_setopt( PvmRoute, PvmAllowDirect );
			}

			else
			{
				if ( vflag ) errprintf( "Direct routing\n" );

				info = pvm_setopt( PvmRoute, PvmRouteDirect );
			}

			start = start_size;
			end = max_size;
			step = incr;

			tag = BANDWIDTH;
		}

		CKERR( info, "pvm_setopt", NOHOST, result = FAILED );

		for ( i=start ; i < end ; i += step )
		{
			Total_time = 0;

			for ( j=0 ; j < TRIALS ; j++ )
			{
				if ( vflag )
					gettimeofday( &t1, NULL );

				if ( master_send( &tid, 1, tag, send_mode, data_format,
						(void *) outbuf, i, PVM_INT,
						hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
				{
					result = FAILED;
					continue;
				}

				/****** Busy loop ? *****/

				for ( count=0 ; count < BUSY_LOOP_PERF ; count++ )
					if ( pvm_probe( tid, tag ) > 0 )
						break;

				if ( count == BUSY_LOOP_PERF )
				{
					errprintf( "\tTimeout expired from host %s\n",
							hostpool[ array_of_index[ length - 1 ] ] );
					result = FAILED;
					continue;
				}

				if ( master_recv( tid, tag, recv_mode,
						(void *) inbuf, i, PVM_INT, (int *) NULL,
						hostpool[ array_of_index[ length - 1 ] ] ) < 0 )
				{
					result = FAILED;
					continue;
				}

				if ( vflag )
				{
					gettimeofday( &t2, NULL );

					Total_time += TIME_ELAPSED;
				}
			}

			if ( vflag )
			{
				if ( !b )
				{
					Average_time =
							( (double) Total_time / (double) TRIALS );

					errprintf( "Average Latency = %1.3f msec\n\n",
							1000 * Average_time / 2 );
					errprintf( "Message Size (bytes) | " );
					errprintf( "Bandwith (Mbyte/sec)\n" );
				}

				else
				{
					Total_size =
							(double) ( sizeof( int ) * i * TRIALS * 2 )
							/ ( 1024.0 * 1024.0 );

					Average_band = Total_size / Total_time;

					errprintf( "%d\t| %1.3f\n",
							i * sizeof( int ), Average_band );
				}
			}
		}

		if ( vflag ) errprintf( "\n" );
	}

	info = pvm_setopt( PvmRoute, PvmAllowDirect );

	CKERR( info, "pvm_setopt", NOHOST, result = FAILED );

fail:

	pvm_kill( tid );

	return( result );
}


/******************/
/*  pvm_notify()  */
/******************/

int
test_notify( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	struct pvmhostinfo *hostp;

	int		host_tid;
	int		infos;
	int		narch;
	int		nhost;
	int		info;
	int		ack;
	int		tid;

	if ( hostpool[1] != NULL )
	{
		/* Host Add Notify */

		info = pvm_notify( PvmHostAdd, NOTIFY, 1, (int *) NULL );

		CKERR( info, "pvm_notify", NOHOST, return( FAILED ) );

		info = pvm_addhosts( hostpool + 1, 1, &infos );

		CKERR( info, "pvm_addhosts", hostpool[1], return( FAILED ) );

		ack = timeout_wait_message( -1, NOTIFY, -1, VERBOSE );

		if ( ack < 0 )
		{
			errprintf( "\tTimed out waiting for host add ack.\n" );
			return( FAILED );
		}

		if ( master_recv( -1, NOTIFY, RECV,
				(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
			return( FAILED );

		/* Host Delete Notify */

		info = pvm_config( &nhost, &narch, &hostp );

		CKERR( info, "pvm_config", NOHOST, return( FAILED ) );

		host_tid = hostp[1].hi_tid;

		info = pvm_notify( PvmHostDelete, NOTIFY, 1, &host_tid );

		CKERR( info, "pvm_notify", NOHOST, return( FAILED ) );

		info = pvm_delhosts( hostpool + 1, 1, &infos );

		CKERR( info, "pvm_delhosts", hostpool[1], return( FAILED ) );

		ack = timeout_wait_message( -1, NOTIFY, -1, VERBOSE );

		if ( ack < 0 )
		{
			errprintf( "\tTimed out waiting for host delete ack.\n" );
			return( FAILED );
		}

		if ( master_recv( -1, NOTIFY, RECV,
				(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
			return( FAILED );

		/* Verify Host Add Notify Count */

		info = pvm_addhosts( hostpool + 1, 1, &infos );

		CKERR( info, "pvm_addhosts", hostpool[1], return( FAILED ) );

		ack = timeout_wait_message( -1, NOTIFY, -1, SILENT );

		if ( ack >= 0 )
		{
			errprintf( "\tError:  Host Add Notify Not Removed " );
			errprintf( "After Count Expired.\n" );
			
			return( FAILED );
		}

		info = pvm_delhosts( hostpool + 1, 1, &infos );

		CKERR( info, "pvm_delhosts", hostpool[1], return( FAILED ) );
	}

	/* Task Exit Notify */

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskDefault, NOHOST, 1, &tid )
			< 0 )
	{
		return( FAILED );
	}

	info = pvm_notify( PvmTaskExit, NOTIFY, 1, &tid );

	CKERR( info, "pvm_notify", NOHOST, return( FAILED ) );

	if ( master_send( &tid, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0, NOHOST ) < 0 )
		return( FAILED );

	ack = timeout_wait_message( -1, ACKNOWLEDGE, -1, VERBOSE );

	if ( ack < 0 )
	{
		errprintf( "\tSlave task did not send acknowledgement.\n" );
		return( FAILED );
	}

	if ( master_recv( tid, ACKNOWLEDGE, RECV,
			(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
		return( FAILED );

	ack = timeout_wait_message( -1, NOTIFY, -1, VERBOSE );

	if ( ack < 0 )
	{
		errprintf( "\tDidn't receive the notification in time.\n" );
		return( FAILED );
	}

	if ( master_recv( -1, NOTIFY, RECV,
			(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
		return( FAILED );

	return( PASSED );
}


/*******************/
/*  Context Tests  */
/*******************/

int
test_context( test_num, array_of_index, length, args )
int     test_num;
int     *array_of_index;
int     length;
int     *args;
{
	int		data[3];
	int		ctx[3];

	int		*tids;

	int		result = PASSED;
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		info;
	int		ack;
	int		tmp;
	int		i, j;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	for( i=0 ; i < length ; i++ )
	{
		if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
				PvmTaskHost, hostpool[ array_of_index[i] ],
				1, &(tids[i]) )
				< 0 )
		{
			result = FAILED;
			break;
		}

		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			break;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, recv_mode,
				(void *) NULL, 0, 0, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			break;
		}
	}

	if ( result == FAILED )
		goto fail;

	for ( j=0 ; j < 3 ; j++ )
		ctx[j] = -1;

	ctx[0] = pvm_getcontext();
	CKERR( ctx[0], "pvm_getcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	ctx[1] = pvm_newcontext();
	CKERR( ctx[1], "pvm_newcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	ctx[2] = pvm_newcontext();
	CKERR( ctx[2], "pvm_newcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	if ( master_send( tids, length, CONTEXT, send_mode, data_format,
			(void *) ctx, 3, PVM_INT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for ( j=0 ; j < 3 ; j++ )
	{
		info = pvm_setcontext( ctx[j] );
		CKERR( info, "pvm_setcontext", NOHOST,
				{ result = FAILED; goto fail; } );

		tmp = 111 * ( j + 1 );

		if ( master_send( tids, length, CONTEXT, send_mode, data_format,
				(void *) &tmp, 1, PVM_INT, NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}
	}

	info = pvm_setcontext( ctx[0] );
	CKERR( info, "pvm_setcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	for ( i=0 ; i < length ; i++ )
	{
		ack = timeout_wait_message( tids[i], CONTEXT,
				array_of_index[i], VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], CONTEXT, recv_mode,
				(void *) data, 3, PVM_INT, (int *) NULL,
				hostpool[ array_of_index[i] ] ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		for ( j=0 ; j < 3 ; j++ )
		{
			if ( data[j] != ( ( data[j] / 111 ) * 111 ) )
			{
				errprintf( "\tGarbled Message:  data[%d] = %d\n",
						j, data[j] );

				result = FAILED;
				continue;
			}

			if ( data[j] != 111 * ( 3 - j ) )
			{
				errprintf( "\tContext Order Mismatch:  " );
				errprintf( "%d received instead of %d\n",
						data[j], 111 * ( 3 - j ) );

				result = FAILED;
			}
		}
	}

fail:

	if ( ctx[0] >= 0 )
	{
		pvm_setcontext( ctx[0] );
		CKERR( info, "pvm_setcontext", NOHOST, result = FAILED );
	}

	for ( j=1 ; j < 3 ; j++ )
	{
		if ( ctx[j] > 0 )
		{
			pvm_freecontext( ctx[j] );
			CKERR( info, "pvm_freecontext", NOHOST, result = FAILED );
		}
	}

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/********************/
/*  Siblings Tests  */
/********************/

int
test_siblings( test_num, array_of_index, length, args )
int     test_num;
int     *array_of_index;
int     length;
int     *args;
{
	int		*tids;
	int		*sibs;

	int		result = PASSED;
	int		factor;
	int		info;
	int		ack;
	int		num;
	int		i, j;

	factor	=	args[0];

	num = length * factor;

	tids = (int *) malloc( num * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	sibs = (int *) malloc( num * sizeof( int ) );
	if ( !memcheck( sibs, "TID Array" ) )
	{
		free( tids );
		return( FAILED );
	}

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskDefault, NOHOST, num, tids ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	for( i=0 ; i < num ; i++ )
	{
		/* Acknoledgement from slave */

		ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
				-1, VERBOSE );

		if ( ack < 0 )
		{
			result = FAILED;
			continue;
		}

		if ( master_recv( tids[i], ACKNOWLEDGE, RECV,
				(void *) sibs, num, PVM_INT, (int *) NULL,
				NOHOST ) < 0 )
		{
			result = FAILED;
			goto fail;
		}

		for ( j=0 ; j < num ; j++ )
		{
			if ( sibs[j] != tids[j] )
			{
				errprintf( "\tSiblings TID Mismatch:  " );
				errprintf( "%d received instead of %d\n",
						sibs[j], tids[j] );

				result = FAILED;
			}
		}
	}

fail:

	for ( i=0 ; i < num ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );
	free( sibs );

	return( result );
}


/***************************/
/*  Message Handler Tests  */
/***************************/

int
dummy_handler( bufid )
int bufid;
{
	int		msgtag;
	int		nbytes;
	int		info;
	int		src;

	info = pvm_bufinfo( bufid, &nbytes, &msgtag, &src );

	CKERR( info, "dummy_handler()/pvm_bufinfo", NOHOST,
			return( FAILED ) );

	errprintf( "Invoked dummy_handler()!!! " );
	errprintf( "(nbytes=%d msgtag=%d src=0x%x)\n",
			nbytes, msgtag, src );

	info = pvm_freebuf( bufid );

	CKERR( info, "dummy_handler()/pvm_freebuf", NOHOST,
			return( FAILED ) );

	return( PASSED );
}

int
test_mhf( test_num, array_of_index, length, args )
int     test_num;
int     *array_of_index;
int     length;
int     *args;
{
	int		tag[3];
	int		ctx[3];

	int		*tids;

	int		result = PASSED;
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		i, j, k;
	int		mhid;
	int		info;
	int		ack;
	int		tmp;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	if ( send_mode == PSEND || recv_mode == PRECV
			|| data_format != PvmDataDefault )
	{
		if ( check_multi_arch() )
			return( NOT_POSSIBLE );
	}

	tids = (int *) malloc( length * sizeof( int ) );
	if ( !memcheck( tids, "TID Array" ) )
		return( FAILED );

	for( i=0 ; i < length ; i++ )
		tids[i] = 0;

	/* Create Some Contexts for Slaves */

	ctx[0] = pvm_getcontext();
	CKERR( ctx[0], "pvm_getcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	ctx[1] = pvm_newcontext();
	CKERR( ctx[1], "pvm_newcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	ctx[2] = pvm_newcontext();
	CKERR( ctx[2], "pvm_newcontext", NOHOST,
			{ result = FAILED; goto fail; } );

	/* Set Local Message Tags Array */

	tag[0] = MHFTAG0;
	tag[1] = MHFTAG1;
	tag[2] = MHFTAG2;

	/*** Incorrect Cases ***/

	/* All Wildcards */

	info = pvm_addmhf( -1, -1, -1, dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	/* (Better Delete It If It Worked!) */
	if ( info >= 0 )
	{
		info = pvm_delmhf( info );
		CKERR( info, "pvm_delmhf", NOHOST, result = FAILED );
	}

	/* Duplicate Entry */

	mhid = pvm_addmhf( 111, 222, ctx[0], dummy_handler );

	CKERR( mhid, "pvm_addmhf", NOHOST, result = FAILED );

	info = pvm_addmhf( 111, 222, ctx[0], dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( -1, 222, ctx[0], dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( 111, -1, ctx[0], dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( 111, 222, -1, dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( -1, -1, ctx[0], dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( 111, -1, -1, dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_addmhf( -1, -1, -1, dummy_handler );

	EXPECTED( "pvm_addmhf", PvmExists, info, NOHOST,
			result = FAILED );

	info = pvm_delmhf( mhid );
	CKERR( info, "pvm_delmhf", NOHOST, result = FAILED );

	/* Spawn Slaves */

	if ( spawn_slave( test_num, DEFAULT_SLAVE_ARGS,
			PvmTaskDefault, NOHOST, length, tids ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* Send Ctxts to Slaves */

	if ( master_send( tids, length, ACKNOWLEDGE, send_mode, data_format,
			(void *) ctx, 3, PVM_INT, NOHOST ) < 0 )
	{
		result = FAILED;
		goto fail;
	}

	/* Send Messages in Context */

	for ( k=0 ; k < 6 ; k++ )
	{
		/* Make Sure Slaves are Ready */

		info = pvm_setcontext( ctx[0] );
		CKERR( info, "pvm_setcontext", NOHOST,
				{ result = FAILED; goto fail; } );

		for( i=0 ; i < length ; i++ )
		{
			/* Acknoledgement from slave */

			ack = timeout_wait_message( tids[i], ACKNOWLEDGE,
					-1, VERBOSE );

			if ( ack < 0 )
			{
				result = FAILED;
				goto fail;
			}

			if ( master_recv( tids[i], ACKNOWLEDGE, recv_mode,
					(void *) NULL, 0, 0, (int *) NULL, NOHOST ) < 0 )
			{
				result = FAILED;
				goto fail;
			}
		}

		for ( j=0 ; j < 3 ; j++ )
		{
			info = pvm_setcontext( ctx[j] );
			CKERR( info, "pvm_setcontext", NOHOST,
					{
						errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
								k, j, ctx[j] );
						result = FAILED;
						goto fail;
					} );

			tmp = 111 * ( j + 1 );

			if ( master_send( tids, length, tag[j],
					send_mode, data_format,
					(void *) &tmp, 1, PVM_INT, NOHOST ) < 0 )
			{
				errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
						k, j, ctx[j] );
				result = FAILED;
				goto fail;
			}
		}

		for ( j=0 ; j < 3 ; j++ )
		{
			info = pvm_setcontext( ctx[ 2 - j ] );
			CKERR( info, "pvm_setcontext", NOHOST,
					{
						errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
								k,  2 - j, ctx[ 2 - j ] );
						result = FAILED;
						goto fail;
					} );

			/* Responses from slave */

			for ( i=0 ; i < length ; i++ )
			{
				ack = timeout_wait_message( tids[i], tag[ 2 - j ],
						-1, VERBOSE );

				if ( ack < 0 )
				{
					errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
							k,  2 - j, ctx[ 2 - j ] );
					result = FAILED;
					goto fail;
				}

				if ( master_recv( tids[i], tag[ 2 - j ], recv_mode,
						(void *) &tmp, 1, PVM_INT, (int *) NULL,
						NOHOST ) < 0 )
				{
					errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
							k,  2 - j, ctx[ 2 - j ] );
					result = FAILED;
					goto fail;
				}

				if ( tmp != 111 * ( 3 - j ) * ( 3 - j ) )
				{
					errprintf( "\tMessage Handler Order Mismatch:  " );
					errprintf( "%d received instead of %d\n",
							tmp, 111 * ( 3 - j ) * ( 3 - j ) );
					errprintf( "\t(case k=%d, ctx[%d]=%d)\n",
							k,  2 - j, ctx[ 2 - j ] );

					result = FAILED;
				}
			}
		}
	}

fail:

	for ( i=0 ; i < length ; i++ )
	{
		if ( tids[i] > 0 )
			pvm_kill( tids[i] );
	}

	free( tids );

	return( result );
}


/*******************************/
/*  Head to Head Test Handler  */
/*******************************/

int
test_head_head( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;
	int		ret;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	ret = head_head( test_num, array_of_index[ length - 1 ],
			is_empty, send_mode, recv_mode, route, data_format );

	return( ret );
}


/***************************/
/*  Triangle Test Handler  */
/***************************/

int
test_triangle( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;
	int		a, b;
	int		ret;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	if ( length == 1 )
	{
		a = array_of_index[ length - 1 ];
		b = a;
	}

	else
	{
		a = array_of_index[ length - 2 ];
		b = array_of_index[ length - 1 ];
	}

	ret = triangle( test_num, a, b,
			is_empty, send_mode, recv_mode, route, data_format );

	return( ret );
}


/****************************/
/*  Funneling Test Handler  */
/****************************/

int
test_funneling( test_num, array_of_index, length, args )
int		test_num;
int		*array_of_index;
int		length;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;
	int		ret;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	ret = funneling( test_num, array_of_index, length,
			is_empty, send_mode, recv_mode, route, data_format );

	return( ret );
}


/**********************/
/*  Spawn Test Slave  */
/**********************/

int
spawn_slave( test_num, slave_args, flags, where, ntasks, tidp )
int		test_num;
int		*slave_args;
int		flags;
char	*where;
int		ntasks;
int		*tidp;
{
	char	*args[ NUM_ARGS + 2 ];

	int *argptr;

	int info;
	int i;

	args[0] = test_str( test_num );

	if ( slave_args != NULL )
		argptr = slave_args;
	
	else
		argptr = module_list[ test_num ].proc_slave_args;

	for ( i=0 ; i < NUM_ARGS ; i++ )
		args[ i + 1 ] = test_str( argptr[i] );
	
	args[ NUM_ARGS + 1 ] = (char *) NULL;

	info = pvm_spawn( TEST_SLAVE, args, flags | slave_spawn_flags,
			where, ntasks, tidp );

	CKERR( info, "pvm_spawn", where, NOHANDLER );

	if ( info < ntasks )
	{
		errprintf( "\tError in spawn_slave():  " );
		errprintf( "%d of %d tasks correctly spawned.\n",
				info, ntasks );

		for ( i=0 ; i < ntasks ; i++ )
			CKERR( tidp[i], "pvm_spawn", where, NOHANDLER ); 

		return( FAILED );
	}

	for ( i=0 ; i < NUM_ARGS + 1 ; i++ )
		free( args[i] );

	return( info );
}


/****************************/
/*  Send / Recv Variations  */
/****************************/

int
master_send( dst, dstlen, tag, send_mode, data_format,
		buf, len, type, host )
int		*dst;
int		dstlen;
int		tag;
int		send_mode;
int		data_format;
void	*buf;
int		len;
int		type;
char	*host;
{
	int		info;
	int		i;

	switch ( send_mode )
	{
		case SEND:
			if ( master_pack_by_type( data_format,
					buf, len, 1, type, PACK, host ) == FAILED )
				return( FAILED );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_send( dst[i], tag );
				CKERR( info, "pvm_send", host, return( FAILED ) );
			}
			break;

		case SENDF:
			if ( master_pack_by_type( data_format,
					buf, len, 1, type, PACKF, host ) == FAILED )
				return( FAILED );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_send( dst[i], tag );
				CKERR( info, "pvm_send", host, return( FAILED ) );
			}
			break;

		case MCAST:
			if ( master_pack_by_type( data_format,
					buf, len, 1, type, PACK, host ) == FAILED )
				return( FAILED );
			info = pvm_mcast( dst, dstlen, tag );
			CKERR( info, "pvm_mcast", host, return( FAILED ) );
			break;

		case PSEND:
			type = ( len > 0 ) ? type : ( ( type ) ? type : PVM_INT );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_psend( dst[i], tag,
						(char *) buf, len, type );
				CKERR( info, "pvm_psend", host, return( FAILED ) );
			}
			break;

		default:
		{
			errprintf(
				"Error in master_send():  Send Mode %d Unknown!\n",
				send_mode );

			return( FAILED );
		}
	}

	return( PASSED );
}


int
master_recv( src, tag, recv_mode, buf, len, type, amsg, host )
int		src;
int		tag;
int		recv_mode;
void	*buf;
int		len;
int		type;
int		*amsg;
char	*host;
{
	int		*atid, *atag, *alen;
	int		bufid;
	int		info;

	switch ( recv_mode )
	{
		case RECV:
			bufid = pvm_recv( src, tag );
			CKERR( bufid, "pvm_recv", host, return( FAILED ) );
			if ( bufid > 0 && len > 0 )
			{
				if ( master_unpack_by_type( buf, len, 1, type,
						UNPACK, host ) == FAILED )
					return( FAILED );
			}
			break;

		case RECVF:
			bufid = pvm_recv( src, tag );
			CKERR( bufid, "pvm_recv", host, return( FAILED ) );
			if ( bufid > 0 && len > 0 )
			{
				if ( master_unpack_by_type( buf, len, 1, type,
						UNPACKF, host ) == FAILED )
					return( FAILED );
			}
			break;

		case PRECV:
			type = ( len > 0 ) ? type : ( ( type ) ? type : PVM_INT );
			if ( amsg ) {
				atid = &(amsg[0]);
				atag = &(amsg[1]);
				alen = &(amsg[2]);
			} else
				atid = atag = alen = (int *) NULL;
			info = pvm_precv( src, tag, (char *) buf, len, type,
					atid, atag, alen );
			CKERR( info, "pvm_precv", host, return( FAILED ) );
			bufid = 0;
			break;

		case NRECV:
			while ( !(bufid = pvm_nrecv( src, tag )) );
			CKERR( bufid, "pvm_nrecv", host, return( FAILED ) );
			if ( bufid > 0 && len > 0 )
			{
				if ( master_unpack_by_type( buf, len, 1, type,
						UNPACK, host ) == FAILED )
					return( FAILED );
			}
			break;

		case TRECV:
			while ( !(bufid = pvm_trecv( src, tag, &TRECV_TM )) );
			CKERR( bufid, "pvm_trecv", host, return( FAILED ) );
			if ( bufid > 0 && len > 0 )
			{
				if ( master_unpack_by_type( buf, len, 1, type,
						UNPACK, host ) == FAILED )
					return( FAILED );
			}
			break;

		default:
		{
			errprintf(
				"Error in master_recv():  Recv Mode %d Unknown!\n",
				recv_mode );

			return( FAILED );
		}
	}

	return( bufid );
}


int
master_pack_by_type( data_format, buf, len, stride, type, ptype, host )
int		data_format;
void	*buf;
int		len;
int		stride;
int		type;
int		ptype;
char	*host;
{
	int		info;

	if ( ptype != PACKF )
	{
		info = pvm_initsend( data_format );
		CKERR( info, "pvm_initsend", host, return( FAILED ) );
	}

	if ( len <= 0 )
	{
		if ( ptype == PACKF )
		{
			info = pvm_packf( "%+", data_format );
			CKERR( info, "pvm_packf", host, return( FAILED ) );
		}

		return( PASSED );
	}

	switch ( type )
	{
		case PVM_STR:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %s", data_format, (char *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkstr( (char *) buf );
				CKERR( info, "pvm_pkint", host, return( FAILED ) );
			}
			break;

		case PVM_BYTE:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*c", data_format,
						len, (char *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkbyte( (char *) buf, len, stride );
				CKERR( info, "pvm_pkbyte", host, return( FAILED ) );
			}
			break;

		case PVM_SHORT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*hd", data_format,
						len, (short *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkshort( (short *) buf, len, stride );
				CKERR( info, "pvm_pkshort", host, return( FAILED ) );
			}
			break;

		case PVM_INT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*d", data_format,
						len, (int *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkint( (int *) buf, len, stride );
				CKERR( info, "pvm_pkint", host, return( FAILED ) );
			}
			break;

		case PVM_FLOAT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*f", data_format,
						len, (float *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkfloat( (float *) buf, len, stride );
				CKERR( info, "pvm_pkfloat", host, return( FAILED ) );
			}
			break;

		case PVM_CPLX:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*x", data_format,
						len, (float *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkcplx( (float *) buf, len, stride );
				CKERR( info, "pvm_pkcplx", host, return( FAILED ) );
			}
			break;

		case PVM_DOUBLE:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*lf", data_format,
						len, (double *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkdouble( (double *) buf, len, stride );
				CKERR( info, "pvm_pkdouble", host, return( FAILED ) );
			}
			break;

		case PVM_DCPLX:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*lx", data_format,
						len, (double *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkdcplx( (double *) buf, len, stride );
				CKERR( info, "pvm_pkdcplx", host, return( FAILED ) );
			}
			break;

		case PVM_LONG:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ld", data_format,
						len, (long *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pklong( (long *) buf, len, stride );
				CKERR( info, "pvm_pklong", host, return( FAILED ) );
			}
			break;

		case PVM_USHORT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*hd", data_format,
						len, (short *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkushort( (unsigned short *) buf, len,
						stride );
				CKERR( info, "pvm_pkushort", host, return( FAILED ) );
			}
			break;

		case PVM_UINT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ud", data_format,
						len, (unsigned int *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkuint( (unsigned int *) buf, len, stride );
				CKERR( info, "pvm_pkuint", host, return( FAILED ) );
			}
			break;

		case PVM_ULONG:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ld", data_format,
						len, (long *) buf );
				CKERR( info, "pvm_packf", host, return( FAILED ) );
			} else {
				info = pvm_pkulong( (unsigned long *) buf, len,
						stride );
				CKERR( info, "pvm_pkulong", host, return( FAILED ) );
			}
			break;

		default:
			errprintf( "\tError in master_pack_by_type():  " );
			errprintf( "Unknown type = %d\n", type );
			return( FAILED );
	}

	return( PASSED );
}


int
master_unpack_by_type( buf, len, stride, type, ptype, host )
char	*buf;
int		len;
int		stride;
int		type;
int		ptype;
char	*host;
{
	int		info;

	switch ( type )
	{
		case PVM_STR:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%s", (char *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkstr( (char *) buf );
				CKERR( info, "pvm_upkstr", host, return( FAILED ) );
			}
			break;

		case PVM_BYTE:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*c", len, (char *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkbyte( (char *) buf, len, stride );
				CKERR( info, "pvm_upkbyte", host, return( FAILED ) );
			}
			break;

		case PVM_SHORT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*hd", len, (short *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkshort( (short *) buf, len, stride );
				CKERR( info, "pvm_upkshort", host, return( FAILED ) );
			}
			break;

		case PVM_INT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*d", len, (int *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkint( (int *) buf, len, stride );
				CKERR( info, "pvm_upkint", host, return( FAILED ) );
			}
			break;

		case PVM_FLOAT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*f", len, (float *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkfloat( (float *) buf, len, stride );
				CKERR( info, "pvm_upkfloat", host, return( FAILED ) );
			}
			break;

		case PVM_CPLX:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*x", len, (float *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkcplx( (float *) buf, len, stride );
				CKERR( info, "pvm_upkcplx", host, return( FAILED ) );
			}
			break;

		case PVM_DOUBLE:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*lf", len, (double *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkdouble( (double *) buf, len, stride );
				CKERR( info, "pvm_upkdouble", host, return( FAILED ) );
			}
			break;

		case PVM_DCPLX:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*lx", len, (double *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkdcplx( (double *) buf, len, stride );
				CKERR( info, "pvm_upkdcplx", host, return( FAILED ) );
			}
			break;

		case PVM_LONG:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ld", len, (long *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upklong( (long *) buf, len, stride );
				CKERR( info, "pvm_upklong", host, return( FAILED ) );
			}
			break;

		case PVM_USHORT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*hd", len, (short *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkushort( (unsigned short *) buf, len,
						stride );
				CKERR( info, "pvm_upkushort", host, return( FAILED ) );
			}
			break;

		case PVM_UINT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ud", len, (unsigned int *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkuint( (unsigned int *) buf, len, stride );
				CKERR( info, "pvm_upkuint", host, return( FAILED ) );
			}
			break;

		case PVM_ULONG:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ld", len, (long *) buf );
				CKERR( info, "pvm_unpackf", host, return( FAILED ) );
			} else {
				info = pvm_upkulong( (unsigned long *) buf, len,
						stride );
				CKERR( info, "pvm_upkulong", host, return( FAILED ) );
			}
			break;

		default:
			errprintf( "\tError in master_unpack_by_type():  " );
			errprintf( "Unknown type = %d\n", type );
			return( FAILED );
	}

	return( PASSED );
}

