
static char rcsid[] =
	"$Id: test_list_slave.c,v 1.29 2001/02/05 22:58:26 pvmsrc Exp $";

#include "pvm_test.h"
#include "globals.h"


void
test_slave_null( test_num, args )
int		test_num;
int		*args;
{}


void
test_slave_ack( test_num, args )
int		test_num;
int		*args;
{
	int		recv_mode;
	int		send_mode;
	int		block;
	int		wait;

	wait		=	args[0];
	block		=	args[1];
	send_mode	=	args[2];
	recv_mode	=	args[3];

	if ( wait == WAIT )
	{
		slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
				(void *) NULL, 0, 0 );
	}

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, PvmDataDefault,
			(void *) NULL, 0, 0 );

	if ( block == BLOCK )
		slave_recv_block();
		
	else
		pvm_exit();
}


void
test_slave_group( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		send_mode;
	int		block;
	int		info;

	block		=	args[0];
	send_mode	=	args[1];

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, PvmDataDefault,
			(void *) NULL, 0, 0 );

	if ( block == BLOCK )
		slave_recv_block();
	
	else
	{
		info = pvm_lvgroup( group );
		CKERR_FP( stderr, info, "pvm_lvgroup", NOHOST, exit( 1 ) );

		pvm_exit();
	}
}


void
test_parent_slave( test_num, args )
int		test_num;
int		*args;
{
	int		send_mode;

	send_mode	=	args[0];

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, PvmDataDefault,
			(void *) &MYPARENT, 1, PVM_INT );

	slave_recv_block();
}


void
sendsig_handler( sig )
int		sig;
{
	slave_send( &MYPARENT, 1, SIGNAL, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );
}


void
test_sendsig_slave( test_num, args )
int		test_num;
int		*args;
{
	sigset_t set;

	sigfillset( &set );
	sigdelset( &set, SIGINT );
	signal( SIGINT, sendsig_handler );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	sigsuspend( &set );

	slave_recv_block();
}


void
test_packing_slave( test_num, args )
int		test_num;
int		*args;
{
	void	*data_in;

	int		data_format;
	int		nelems;
	int		send_mode;
	int		recv_mode;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	nelems		=	args[2];
	data_format	=	args[3];

	data_in = (void *) malloc( ( (unsigned) ( nelems * 2 ) )
			* sizeof( double ) );
	if ( !memcheck( data_in, "Packing Data" ) )
		exit( -1 );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	/* CHAR */

	if ( send_mode != PSEND && recv_mode != PRECV )
	{
		slave_recv( MYPARENT, TEST_VALUE, recv_mode,
				data_in, nelems, PVM_STR );

		slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
				data_in, nelems, PVM_STR );
	}

	/* BYTE */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_BYTE );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_BYTE );

	/* INT */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_INT );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_INT );

	/* SHORT */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_SHORT );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_SHORT );

	/* LONG */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_LONG );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_LONG );

	/* U_INT */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_UINT );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_UINT );

	/* U_SHORT */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_USHORT );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_USHORT );

	/* U_LONG */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_ULONG );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_ULONG );

	/* FLOAT */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_FLOAT );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_FLOAT );

	/* DOUBLE */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, nelems, PVM_DOUBLE );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, nelems, PVM_DOUBLE );

	/* CPLX */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, ( nelems == 1 ) ? nelems : nelems / 2, PVM_CPLX );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, ( nelems == 1 ) ? nelems : nelems / 2, PVM_CPLX );

	/* DCPLX */

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			data_in, ( nelems == 1 ) ? nelems : nelems / 2, PVM_DCPLX );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, data_format,
			data_in, ( nelems == 1 ) ? nelems : nelems / 2, PVM_DCPLX );

	slave_recv_block();
}


void
test_bufinfo_slave( test_num, args )
int		test_num;
int		*args;
{
	int		message = 3;

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) &message, 1, PVM_INT );

	slave_recv_block();
}


void
test_mytid_slave( test_num, args )
int		test_num;
int		*args;
{
	int		message;

	message = pvm_mytid();
	CKERR_FP( stderr, message, "pvm_mytid", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) &message, 1, PVM_INT );

	slave_recv_block();
}


void
test_coherency_slave( test_num, args )
int		test_num;
int		*args;
{
	double	message[10];

	int		send_mode;
	int		recv_mode;

	send_mode = args[0];
	recv_mode = args[1];

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_VALUE, recv_mode,
			(void *) message, 10, PVM_DOUBLE );

	slave_send( &MYPARENT, 1, TEST_VALUE, send_mode, PvmDataDefault,
			(void *) message, 10, PVM_DOUBLE );

	pvm_exit();
}


void
test_setsbuf_slave( test_num, args )
int		test_num;
int		*args;
{
	int		message;

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_VALUE, RECV,
			(void *) &message, 1, PVM_INT );

	slave_send( &MYPARENT, 1, TEST_VALUE, SEND, PvmDataDefault,
			(void *) &message, 1, PVM_INT );

	slave_recv_block();
}


void
test_setrbuf_slave( test_num, args )
int		test_num;
int		*args;
{
	int		message;

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	message = pvm_mytid();
	CKERR_FP( stderr, message, "pvm_mytid", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, TEST_VALUE, SEND, PvmDataDefault,
			(void *) &message, 1, PVM_INT );

	slave_recv_block();
}


void
test_bcast_slave( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		info;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_CAST, RECV, (void *) NULL, 0, 0 );

	slave_send( &MYPARENT, 1, TEST_CAST, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	info = pvm_lvgroup( group );
	CKERR_FP( stderr, info, "pvm_lvgroup", NOHOST, exit( 1 ) );

	pvm_exit();
}


void
test_barrier_slave( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		number;
	int		info;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_BARRIER, RECV,
			(void *) &number, 1, PVM_INT );

	info = pvm_barrier( group, number + 1 );
	CKERR_FP( stderr, info, "pvm_barrier", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, TEST_BARRIER, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	info = pvm_lvgroup( group );
	CKERR_FP( stderr, info, "pvm_lvgroup", NOHOST, exit( 1 ) );

	pvm_exit();
}


void
test_reduce_slave( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		*data;

	int		count = 10;
	int		root = -1;
	int		info;
	int		i;

	data = (int *) malloc( count * sizeof( int ) );
	if ( !memcheck( data, "Reduce Data Array" ) )
		exit( -1 );

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_REDUCE, RECV,
			(void *) &root, 1, PVM_INT );

	for ( i=0 ; i < count ; i++ )
		data[i] = i;

	info = pvm_reduce( PvmSum, data, count, PVM_INT, TEST_REDUCE,
			group, root );
	CKERR_FP( stderr, info, "pvm_reduce", NOHOST, exit( 1 ) );

	slave_recv_block();
}


void
test_gather_slave( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		*data;

	int		count = 10;
	int		result;
	int		root;
	int		info;
	int		i;

	data = (int *) malloc( sizeof( int ) * count );
	if ( !memcheck( data, "Gather Data Array" ) )
		exit( -1 );

	for ( i=0 ; i < count ; i++ )
		data[i] = i;

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_GATHER, RECV,
			(void *) &root, 1, PVM_INT );

	info = pvm_gather( &result, data, count, PVM_INT, TEST_GATHER,
			group, root );
	CKERR_FP( stderr, info, "pvm_gather", NOHOST, exit( 1 ) );

	slave_recv_block();
}


void
test_scatter_slave( test_num, args )
int		test_num;
int		*args;
{
	char	group[255];

	int		*data;

	int		count = 5;
	int		result;
	int		root;
	int		info;

	data = (int *) malloc( count * sizeof( int ) );
	if ( !memcheck( data, "Scatter Data Array" ) )
		exit( -1 );

	sprintf( group, "%s%d", GROUP_NAME, test_num );

	info = pvm_joingroup( group );
	CKERR_FP( stderr, info, "pvm_joingroup", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_SCATTER, RECV,
			(void *) &root, 1, PVM_INT );

	info = pvm_scatter( data, &result, count, PVM_INT, TEST_SCATTER,
			group, root );
	CKERR_FP( stderr, info, "pvm_scatter", NOHOST, exit( 1 ) );

	slave_recv( MYPARENT, TEST_SCATTER, RECV, (void *) NULL, 0, 0 );

	slave_send( &MYPARENT, 1, TEST_SCATTER, SEND, PvmDataDefault,
			(void *) data, count, PVM_INT );

	info = pvm_lvgroup( group );
	CKERR_FP( stderr, info, "pvm_lvgroup", NOHOST, exit( 1 ) );

	pvm_exit();
}


void
test_mcast_slave( test_num, args )
int		test_num;
int		*args;
{
	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, TEST_MCAST, RECV, (void *) NULL, 0, 0 );

	slave_send( &MYPARENT, 1, TEST_MCAST, SEND, PvmDataDefault,
			(void *) NULL, 0, 0 );

	pvm_exit();
}


void
test_context_slave( test_num, args )
int		test_num;
int		*args;
{
	int		data[3];
	int		ctx[3];

	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		info;
	int		i;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, CONTEXT, recv_mode,
			(void *) ctx, 3, PVM_INT );

	for ( i=2 ; i >= 0 ; i-- )
	{
		info = pvm_setcontext( ctx[i] );
		CKERR_FP( stderr, info, "pvm_setcontext", NOHOST, exit( 1 ) );

		slave_recv( MYPARENT, CONTEXT, recv_mode,
				(void *) &(data[ 2 - i ]), 1, PVM_INT );
	}

	slave_send( &MYPARENT, 1, CONTEXT, send_mode, data_format,
			(void *) data, 3, PVM_INT );

	slave_recv_block();
}


void
test_siblings_slave( test_num, args )
int		test_num;
int		*args;
{
	int		*sibs;

	int		nsibs;
	int		i;

	nsibs = pvm_siblings( &sibs );
	CKERR_FP( stderr, nsibs, "pvm_siblings", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, SEND, PvmDataDefault,
			(void *) sibs, nsibs, PVM_INT );

	slave_recv_block();
}


int		mhf_data_format;
int		mhf_ctx[3];
int		mhf_count;

int
dummy_handler_slave( bufid, index )
int		bufid;
int		index;
{
	int		old_ctx;
	int		msgtag;
	int		nbytes;
	int		info;
	int		rbf;
	int		src;
	int		tmp;

	/* Decr Recv Count */

	mhf_count++;

	/* Unpack Message */

	info = pvm_bufinfo( bufid, &nbytes, &msgtag, &src );
	CKERR_FP( stderr, info, "pvm_bufinfo", NOHOST, exit( 1 ) );

	rbf = pvm_setrbuf( bufid );

	info = pvm_upkint( &tmp, 1, 1 );
	CKERR_FP( stderr, info, "pvm_upkint", NOHOST, exit( 1 ) );

	info = pvm_freebuf( pvm_setrbuf( rbf ) );
	CKERR_FP( stderr, info, "pvm_freebuf", NOHOST, exit( 1 ) );

	/* Send Reply */

	tmp *= ( index + 1 );

	old_ctx = pvm_setcontext( mhf_ctx[ index ] );
	CKERR_FP( stderr, old_ctx, "pvm_setcontext", NOHOST, exit( 1 ) );

	slave_send( &src, 1, msgtag, SEND, mhf_data_format,
			(void *) &tmp, 1, PVM_INT );

	info = pvm_setcontext( old_ctx );
	CKERR_FP( stderr, info, "pvm_setcontext", NOHOST, exit( 1 ) );

	return( PASSED );
}

int
dummy_handler0_slave( bufid )
int bufid;
{
	return( dummy_handler_slave( bufid, 0 ) );
}

int
dummy_handler1_slave( bufid )
int bufid;
{
	return( dummy_handler_slave( bufid, 1 ) );
}

int
dummy_handler2_slave( bufid )
int bufid;
{
	return( dummy_handler_slave( bufid, 2 ) );
}

void
test_mhf_slave( test_num, args )
int		test_num;
int		*args;
{
	ifp		func[3];

	int		mhid[3];
	int		ctx[3];
	int		tag[3];

	int		send_mode;
	int		recv_mode;
	int		data_format;
	int		s, t, c;
	int		info;
	int		j, k;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	mhf_data_format = data_format;

	func[0] = dummy_handler0_slave;
	func[1] = dummy_handler1_slave;
	func[2] = dummy_handler2_slave;

	tag[0] = MHFTAG0;
	tag[1] = MHFTAG1;
	tag[2] = MHFTAG2;

	slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
			(void *) ctx, 3, PVM_INT );

	mhf_ctx[0] = ctx[0];
	mhf_ctx[1] = ctx[1];
	mhf_ctx[2] = ctx[2];

	for ( k=0 ; k < 6 ; k++ )
	{
		for ( j=0 ; j < 3 ; j++ )
		{
			s = MYPARENT;
			t = tag[j];
			c = ctx[j];

			switch ( k )
			{
				case 0:
					break;
				case 1:
					s = -1;
					break;
				case 2:
					t = -1;
					break;
				case 3:
					c = -1;
					break;
				case 4:
					s = -1;
					t = -1;
					break;
				case 5:
					s = -1;
					c = -1;
					break;
			}

			mhid[j] = pvm_addmhf( s, t, c, func[j] );
		}

		mhf_count = 0;

		slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

		while ( mhf_count < 3 )
		{
			info = pvm_probe( -1, -1 );
			CKERR_FP( stderr, info, "pvm_probe", NOHOST, exit( 1 ) );

			sleep( 1 );
		}

		for ( j=0 ; j < 3 ; j++ )
		{
			info = pvm_delmhf( mhid[j] );
			CKERR_FP( stderr, info, "pvm_delmhf", NOHOST, exit( 1 ) );
		}
	}

	slave_recv_block();
}


void
test_perf_slave( test_num, args )
int		test_num;
int		*args;
{
	int		outbuf[ MAX_MESSAGE_LENGTH ];
	int		inbuf[ MAX_MESSAGE_LENGTH ];
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		data[3];
	int		info;
	int		i, j;

	send_mode	=	args[0];
	recv_mode	=	args[1];
	data_format	=	args[2];

	info = pvm_setopt( PvmRoute, PvmAllowDirect );
	CKERR_FP( stderr, info, "pvm_setopt", NOHOST, exit( 1 ) );

	slave_send( &MYPARENT, 1, ACKNOWLEDGE, send_mode, data_format,
			(void *) NULL, 0, 0 );

	slave_recv( MYPARENT, ACKNOWLEDGE, recv_mode,
			(void *) data, 3, PVM_INT );

	for ( j=0 ; j < TRIALS ; j++ )
	{
		slave_recv( MYPARENT, LATENCY, recv_mode, (void *) NULL, 0, 0 );

		slave_send( &MYPARENT, 1, LATENCY, send_mode, data_format,
				(void *) NULL, 0, 0 );
	}

	for ( i=data[1] ; i < data[0] ; i += data[2] )
	{
		for ( j=0 ; j < TRIALS ; j++ )
		{
			slave_recv( MYPARENT, BANDWIDTH, recv_mode,
					(void *) inbuf, i, PVM_INT );

			slave_send( &MYPARENT, 1, BANDWIDTH, send_mode, data_format,
					(void *) outbuf, i, PVM_INT );
		}
	}

	info = pvm_setopt( PvmRoute, PvmRouteDirect );
	CKERR_FP( stderr, info, "pvm_setopt", NOHOST, exit( 1 ) );

	for ( i=data[1] ; i < data[0] ; i += data[2] )
	{
		for ( j=0 ; j < TRIALS ; j++ )
		{
			slave_recv( MYPARENT, BANDWIDTH, recv_mode,
					(void *) inbuf, i, PVM_INT );

			slave_send( &MYPARENT, 1, BANDWIDTH, send_mode, data_format,
					(void *) outbuf, i, PVM_INT );
		}
	}

	pvm_exit();
}


/*************************************/
/*  Head to Head Slave Test Handler  */
/*************************************/

void
test_head_head_slave( test_num, args )
int		test_num;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	head_head_slave( is_empty, send_mode, recv_mode,
			route, data_format );
}


/*********************************/
/*  Triangle Slave Test Handler  */
/*********************************/

void
test_triangle_slave( test_num, args )
int		test_num;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	triangle_slave( is_empty, send_mode, recv_mode,
			route, data_format );
}


/**********************************/
/*  Funneling Slave Test Handler  */
/**********************************/

void
test_funneling_slave( test_num, args )
int		test_num;
int		*args;
{
	int		data_format;
	int		send_mode;
	int		recv_mode;
	int		is_empty;
	int		route;

	is_empty	=	args[0];
	send_mode	=	args[1];
	recv_mode	=	args[2];
	route		=	args[3];
	data_format	=	args[4];

	funneling_slave( is_empty, send_mode, recv_mode,
			route, data_format );
}


void
slave_send( dst, dstlen, tag, send_mode, data_format, buf, len, type )
int		*dst;
int		dstlen;
int		tag;
int		send_mode;
int		data_format;
void	*buf;
int		len;
int		type;
{
	int		info;
	int		sbf;
	int		i;

	switch ( send_mode )
	{
		case SEND:
			sbf = pvm_setsbuf( 0 );
			slave_pack_by_type( data_format, buf, len, 1, type, PACK );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_send( dst[i], tag );
				CKERR_FP( stderr, info, "pvm_send", NOHOST, exit( 1 ) );
			}
			pvm_setsbuf( sbf );
			break;

		case SENDF:
			sbf = pvm_setsbuf( 0 );
			slave_pack_by_type( data_format, buf, len, 1, type, PACKF );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_send( dst[i], tag );
				CKERR_FP( stderr, info, "pvm_send", NOHOST, exit( 1 ) );
			}
			pvm_setsbuf( sbf );
			break;

		case MCAST:
			sbf = pvm_setsbuf( 0 );
			slave_pack_by_type( data_format, buf, len, 1, type, PACK );
			info = pvm_mcast( dst, dstlen, tag );
			CKERR_FP( stderr, info, "pvm_mcast", NOHOST, exit( 1 ) );
			pvm_setsbuf( sbf );
			break;

		case PSEND:
			type = ( len > 0 ) ? type : ( ( type ) ? type : PVM_INT );
			for ( i=0 ; i < dstlen ; i++ )
			{
				info = pvm_psend( dst[i], tag,
						(char *) buf, len, type );
				CKERR_FP( stderr, info, "pvm_psend",
						NOHOST, exit( 1 ) );
			}
			break;

		default:
			fprintf( stderr,
				"Error in slave_send():  Send Mode %d Unknown!\n",
				send_mode );
	}
}


void
slave_recv( src, tag, recv_mode, buf, len, type )
int		src;
int		tag;
int		recv_mode;
void	*buf;
int		len;
int		type;
{
	char	tmp[255];

	int		bufid;

	switch ( recv_mode )
	{
		case RECV:
			bufid = pvm_recv( src, tag );
			CKERR_FP( stderr, bufid, "pvm_recv", NOHOST, exit( 1 ) );
			if ( len > 0 )
				slave_unpack_by_type( buf, len, 1, type, UNPACK );
			break;

		case RECVF:
			bufid = pvm_recv( src, tag );
			CKERR_FP( stderr, bufid, "pvm_recv", NOHOST, exit( 1 ) );
			if ( len > 0 )
				slave_unpack_by_type( buf, len, 1, type, UNPACKF );
			break;

		case PRECV:
			type = ( len > 0 ) ? type : ( ( type ) ? type : PVM_INT );
			bufid = pvm_precv( src, tag, (char *) buf, len, type,
					(int *) NULL, (int *) NULL, (int *) NULL );
			CKERR_FP( stderr, bufid, "pvm_precv", NOHOST, exit( 1 ) );
			break;

		case NRECV:
			while ( !(bufid = pvm_nrecv( src, tag )) );
			CKERR_FP( stderr, bufid, "pvm_nrecv", NOHOST, exit( 1 ) );
			if ( bufid > 0 && len > 0 )
				slave_unpack_by_type( buf, len, 1, type, UNPACK );
			break;

		case TRECV:
			while ( !(bufid = pvm_trecv( src, tag, &TRECV_TM )) );
			CKERR_FP( stderr, bufid, "pvm_trecv", NOHOST, exit( 1 ) );
			if ( bufid > 0 && len > 0 )
				slave_unpack_by_type( buf, len, 1, type, UNPACK );
			break;

		default:
			fprintf( stderr,
				"Error in slave_recv():  Recv Mode %d Unknown!\n",
				recv_mode );
	}
}


void
slave_recv_block()
{
	int info;
	
	info = pvm_recv( -1, ENDLESS );
	CKERR_FP( stderr, info, "pvm_recv", NOHOST, exit( 1 ) );
}


void
slave_pack_by_type( data_format, buf, len, stride, type, ptype )
int		data_format;
void	*buf;
int		len;
int		stride;
int		type;
int		ptype;
{
	int		info;

	if ( ptype != PACKF )
	{
		info = pvm_initsend( data_format );
		CKERR_FP( stderr, info, "pvm_initsend", NOHOST, exit( 1 ) );
	}

	if ( len <= 0 )
	{
		if ( ptype == PACKF )
		{
			info = pvm_packf( "%+", data_format );
			CKERR_FP( stderr, info, "pvm_packf", NOHOST, exit( 1 ) );
		}

		return;
	}

	switch ( type )
	{
		case PVM_STR:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %s", data_format, (char *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkstr( (char *) buf );
				CKERR_FP( stderr, info, "pvm_pkstr",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_BYTE:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*c", data_format,
						len, (char *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkbyte( (char *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkbyte",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_SHORT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*hd", data_format,
						len, (short *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkshort( (short *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkshort",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_INT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*d", data_format,
						len, (int *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkint( (int *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkint",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_FLOAT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*f", data_format,
						len, (float *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkfloat( (float *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkfloat",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_CPLX:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*x", data_format,
						len, (float *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkcplx( (float *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkcplx",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_DOUBLE:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*lf", data_format,
						len, (double *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkdouble( (double *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkdouble",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_DCPLX:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*lx", data_format,
						len, (double *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkdcplx( (double *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkdcplx",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_LONG:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ld", data_format,
						len, (long *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pklong( (long *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pklong",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_USHORT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*hd", data_format,
						len, (short *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkushort( (unsigned short *) buf, len,
						stride );
				CKERR_FP( stderr, info, "pvm_pkushort",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_UINT:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ud", data_format,
						len, (unsigned int *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkuint( (unsigned int *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_pkuint",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_ULONG:
			if ( ptype == PACKF ) {
				info = pvm_packf( "%+ %*ld", data_format,
						len, (long *) buf );
				CKERR_FP( stderr, info, "pvm_packf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_pkulong( (unsigned long *) buf, len,
						stride );
				CKERR_FP( stderr, info, "pvm_pkulong",
						NOHOST, exit( 1 ) );
			}
			break;

		default:
			printf( "\tError in pack_by_type():  " );
			printf( "Unknown type = %d\n", type );
	}
}


void
slave_unpack_by_type( buf, len, stride, type, ptype )
char	*buf;
int		len;
int		stride;
int		type;
int		ptype;
{
	int		info;

	switch ( type )
	{
		case PVM_STR:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%s", (char *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkstr( (char *) buf );
				CKERR_FP( stderr, info, "pvm_upkstr",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_BYTE:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*c", len, (char *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkbyte( (char *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkbyte",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_SHORT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*hd", len, (short *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkshort( (short *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkshort",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_INT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*d", len, (int *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkint( (int *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkint",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_FLOAT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*f", len, (float *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkfloat( (float *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkfloat",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_CPLX:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*x", len, (float *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkcplx( (float *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkcplx",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_DOUBLE:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*lf", len, (double *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkdouble( (double *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkdouble",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_DCPLX:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*lx", len, (double *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkdcplx( (double *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkdcplx",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_LONG:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ld", len, (long *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upklong( (long *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upklong",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_USHORT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*hd", len, (short *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkushort( (unsigned short *) buf, len,
						stride );
				CKERR_FP( stderr, info, "pvm_upkushort",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_UINT:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ud", len, (unsigned int *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkuint( (unsigned int *) buf, len, stride );
				CKERR_FP( stderr, info, "pvm_upkuint",
						NOHOST, exit( 1 ) );
			}
			break;

		case PVM_ULONG:
			if ( ptype == UNPACKF ) {
				info = pvm_unpackf( "%*ld", len, (long *) buf );
				CKERR_FP( stderr, info, "pvm_unpackf",
						NOHOST, exit( 1 ) );
			} else {
				info = pvm_upkulong( (unsigned long *) buf, len,
						stride );
				CKERR_FP( stderr, info, "pvm_upkulong",
						NOHOST, exit( 1 ) );
			}
			break;

		default:
			printf( "\tError in unpack_by_type():  " );
			printf( "Unknown type = %d\n", type );
	}
}

